package contrabase;

import java.io.IOException;
import java.io.Serializable;
import java.lang.reflect.Field;
import java.lang.reflect.Modifier;
import java.util.HashMap;
import java.util.Iterator;
import java.util.Map;

import contrabase.util.Converter;

/**
 * @author Jong10
 */
public class ObjectContainer implements Serializable {
	private static final long serialVersionUID = 1L;
	
	private String className;
	private String key;
	private int repositoryNo;
	private String rid;
	
	public String getClassName() {
		return className;
	}
	
	public void setClassName(String className) {
		this.className = className;
	}

	public String getKey() {
		return key;
	}
	
	public void setKey(String key) {
		this.key = key;
	}
	
	private Map<String, ObjectContainerField> fields = new HashMap<String, ObjectContainerField>();
	
	
	/**
	 * 새 'ObjectContainerField' 개체를 추가한다.
	 * @param field
	 */
	public void add(ObjectContainerField field) {
		fields.put(field.getFieldName(), field);
	}
	
	/**
	 * @return
	 */
	public Iterator<ObjectContainerField> iterator() {
		return fields.values().iterator();
	}
	
	/**
	 * @throws IOException
	 * @throws ClassNotFoundException
	 */
	public ObjectContainer() {
		this.className = "";
	}
	
	/**
	 * 'Serializable'을 구현한 객체의 인스턴스를 매개변수로 받는다.
	 * 해당 객체의 필드를 분해해서, ObjectContainer 안에 보관한다.
	 * 
	 * @param object
	 * @throws ClassNotFoundException 
	 * @throws IOException 
	 */
	public ObjectContainer(String key, Serializable object) throws IOException, ClassNotFoundException {
		Class<?> c = object.getClass();
		this.className = c.getName();
		this.key = key;
		this.repositoryNo = 0;
		
		do {
			extractFields(object, c);
			c = c.getSuperclass();
			// 상속한 모든 부모 클래스의 필드도 모두 담는다. 
		} while (isDeclaredClass(c));
	}
	
	/**
	 * 사용자가 정의한 클래스인지를 판단한다.
	 * @param c
	 * @return
	 */
	private boolean isDeclaredClass(Class<?> c) {
		return c != null && c.getName().equals("java.lang.Object") == false;
	}

	/**
	 * 해당 클래스의 각 필드를 뒤져서, fields에 담는다.
	 * @param object
	 * @param theClass
	 * @throws IOException
	 */
	private void extractFields(Serializable object, Class<?> theClass) throws IOException {
		for (Field f : theClass.getDeclaredFields()) {
			// 저장해야할 필드인지 확인하고,
			if (isNeedfulField(f)) {
				String typeName = f.getType().getName();
				String fieldName = f.getName();
				
				Serializable value = null;
				
				// 필드의 값들을, 'Object'로 가져온다.
				try {
					f.setAccessible(true);	// private, final 등에도 접근 가능하게 한다.
					value = (Serializable)getFieldValue(object, f);
				} catch (IllegalArgumentException e) {
					e.printStackTrace();
				} catch (IllegalAccessException e) {
					e.printStackTrace();
				}
				
				// 해당 필드의 이름과, 타입이름, 값을 넣는다.
				fields.put(fieldName, new ObjectContainerField(fieldName, typeName, value));
			}
		}
	}
	
	private boolean isNeedfulField(Field f) {
		// XXX: final, static 은 어떻게 해야하나?? 에러가 발생하는데.
		return isNotTransient(f) && isNotSerialVersionUID(f);
	}

	/**
	 * 이 객체의 필드들을, 해당 정보에 맞춰서, Object 로 재구성해서 리턴한다.
	 * 이 클래스가 동작하는 JVM에, className에 담긴 이름의 클래스가 로딩되어 있어야 한다.
	 * 
	 * @return ObjectContainer 안에 담겨있던 객체를 원래대로 복구해준다.
	 * @throws ClassNotFoundException
	 * @throws IOException 
	 */
	public Serializable toObject() throws ClassNotFoundException, IOException {
		Serializable object = null;
		Class<?> c = Class.forName(this.className);
		
		try {
			// XXX: 기본 생성자가 없으면, 에러가 난다. -_-;;;;;
			object = (Serializable)c.newInstance();
		} catch (InstantiationException e) {
			e.printStackTrace();
		} catch (IllegalAccessException e) {
			e.printStackTrace();
		}
		
		do {
			restoreFields(object, c);
			c = c.getSuperclass();
		} while (isDeclaredClass(c));
		
		return object;
	}


	/**
	 * 필드들을 복구 한다.
	 * @param targetObject
	 * @param c
	 * @throws IOException
	 * @throws ClassNotFoundException
	 */
	private void restoreFields(Serializable targetObject, Class<?> c)
			throws IOException, ClassNotFoundException {
		for (Field f : c.getDeclaredFields()) {
			
			// 저장해야 할 필드일 경우에만..
			if (isNeedfulField(f)) {
				ObjectContainerField theField = fields.get(f.getName());
				Serializable value = theField.getValue();
				f.setAccessible(true);
				
				try {
					setFieldValue(targetObject, f, value);
				} catch (IllegalArgumentException e) {
					e.printStackTrace();
				} catch (IllegalAccessException e) {
					e.printStackTrace();
				}
			}
		}
	}
	
	
	/**
	 * @param object
	 * @param field
	 * @param value
	 * @throws IllegalArgumentException
	 * @throws IllegalAccessException
	 * @throws IOException 
	 * @throws ClassNotFoundException 
	 * @throws ClassNotFoundException 
	 */
	private void setFieldValue(Object object, Field field, Serializable value)
	throws IllegalArgumentException, IllegalAccessException, IOException, ClassNotFoundException {
		
		String typeName = field.getType().getName();
		
		// 객체를 primitive 타입으로 변환하고,
		// 'Boxing'된 클래스와 'String'은 객체 자체를 그대로 전달하고,
		// 이외의 객체는 byte[]로 저장되었으므로, 이를 'Object'로 복구한다.
		
		if (typeName.equals("boolean"))
			field.setBoolean(object, (Boolean)value);
		
		else if (typeName.equals("byte"))
			field.setByte(object, (Byte)value);
		
		else if (typeName.equals("char"))
			field.setChar(object, (Character)value);
		
		else if (typeName.equals("double"))
			field.setDouble(object, (Double)value);
		
		else if (typeName.equals("float"))
			field.setFloat(object, (Float)value);
		
		else if (typeName.equals("int"))
			field.setInt(object, (Integer)value);
		
		else if (typeName.equals("long"))
			field.setLong(object, (Long)value);
		
		else if (typeName.equals("short"))
			field.setShort(object, (Short)value);
		
		else if (typeName.equals("java.lang.String")
				|| typeName.equals("java.lang.Boolean")
				|| typeName.equals("java.lang.Byte")
				|| typeName.equals("java.lang.Character")
				|| typeName.equals("java.lang.Double")
				|| typeName.equals("java.lang.Float")
				|| typeName.equals("java.lang.Integer")
				|| typeName.equals("java.lang.Long")
				|| typeName.equals("java.lang.Short"))
			field.set(object, value);
		
		else
			// Byte[]배열을 'Object'로 변환.
			field.set(object, Converter.bytesToObject((byte[])value));
	}
	
	
	/**
	 * obj 객체의 해당 필드를, primitive 값들을 걸러서 값을 Object 로 가져온다.
	 * primitive 타입들은 boxing 되서 리턴된다.
	 * 
	 * @param object
	 * @param field
	 * @return
	 * @throws IllegalArgumentException
	 * @throws IllegalAccessException
	 * @throws ClassNotFoundException 
	 * @throws IOException 
	 */
	private Object getFieldValue(Object object, Field field) 
	throws IllegalArgumentException, IllegalAccessException, IOException {
		
		String typeName = field.getType().getName();
		Object value = null;
		
		// primitive 타입을 해당 메소드로 객체로 바꾸고,
		// 'Boxing'된 클래스와 'String'은 그대로.
		// 이외의 객체는 byte[]로 변환한다.
		
		if (typeName.equals("boolean"))
			value = field.getBoolean(object);
		
		else if (typeName.equals("byte"))
			value = field.getByte(object);
		
		else if (typeName.equals("char"))
			value = field.getChar(object);
		
		else if (typeName.equals("double"))
			value = field.getDouble(object);
		
		else if (typeName.equals("float"))
			value = field.getFloat(object);
		
		else if (typeName.equals("int"))
			value = field.getInt(object);
		
		else if (typeName.equals("long"))
			value = field.getLong(object);
		
		else if (typeName.equals("short"))
			value = field.getShort(object);
		
		else if (typeName.equals("java.lang.String")
				|| typeName.equals("java.lang.Boolean")
				|| typeName.equals("java.lang.Byte")
				|| typeName.equals("java.lang.Character")
				|| typeName.equals("java.lang.Double")
				|| typeName.equals("java.lang.Float")
				|| typeName.equals("java.lang.Integer")
				|| typeName.equals("java.lang.Long")
				|| typeName.equals("java.lang.Short"))
			value = field.get(object);
		
		else
			// 'Object'에서 byte[]로 변환
			value = Converter.objectToBytes((Serializable)field.get(object));
		
		
		return value;
	}
	
	/**
	 * 해당 필드가 serialVersionUID 가 아닌지 검사한다.
	 * serialVersionUID 이면 저장하지 않는다.
	 * 
	 * @param field
	 * @return
	 */
	private boolean isNotSerialVersionUID(Field field) {
		return "serialVersionUID".equals(field.getName()) == false;
	}
	
	/**
	 * 해당 필드가 `transient`가 아닌지 검사한다. `transient`이면 저장하지 않는다. 
	 * 
	 * @param field
	 * @return
	 */
	private boolean isNotTransient(Field field) {
		return (field.getModifiers() & Modifier.TRANSIENT) == 0;
	}
	
	/**
	 * 가지고 있는 필드들의 사이즈를 구한다.
	 * @return 필드들의 사이즈
	 */
	public int size() {
		return fields.size();
	}
	
	/**
	 * 리포지토리 고유 ID를 설정한다.
	 * @return 리포지토리 고유ID
	 */
	public String getRepositoryId(){
		return rid;
	}
	/**
	 * 리포지토리 고유 ID인 RID를 설정한다
	 * @param rid 
	 */
	public void setRepositoryId(String rid){
		this.rid = rid;
	}
	/**
	 * 리포지토리 번호를 가져온다. (RID와는 다른 개념이다.)
	 * @return 리포지토리 번호
	 */
	public int getRepositoryNo() {
		return repositoryNo;
	}
	
	/**
	 * 리포지토리 번호를 설정한다. (RID와는 다른 개념이다.)
	 * @return 리포지토리 번호
	 */
	public void setRepositoryNo(int repositoryNo) {
		this.repositoryNo = repositoryNo;
	}
}
