package contrabase.api;

import java.util.List;
import java.util.Map;

import contrabase.api.Table;
import contrabase.network.ClientSide;
import contrabase.network.ContrabaseNetworkException;
import contrabase.network.MSG;
import contrabase.network.Message;

public class ContrabaseConnector {
	private ClientSide connection;
	private String userId;
	private String password;
	private long clientId = 0;

	/**
	 * 서버와의 새 연결을 만든다.
	 * 
	 * @param ip 서버의 아이피
	 * @param port 서버의 포트
	 * @throws ContrabaseNetworkException
	 */
	public ContrabaseConnector(String ip, int port, String userId, String password)
	throws ContrabaseNetworkException {
		this.userId = userId;
		this.password = password;
		open(ip, port);
	}

	/**
	 * 마스터와 새 연결을 생성한다.
	 * @param ip 마스터의 IP 주소
	 * @param port 마스터의 포트
	 * @throws ContrabaseNetworkException
	 */
	private void open(String ip, int port) throws ContrabaseNetworkException {
		connection = new ClientSide(ip, port);
		
		Message request = new Message();
		request.set(MSG.ACTION, MSG.REQUEST.OPEN_CONNECTION);
		request.set(MSG.CONNECTION.USERNAME, userId);
		request.set(MSG.CONNECTION.PASSWORD, password);
		
		Message response = connection.send(request);
		if (response.isSuccessed()) {
			clientId = (Long)response.get(MSG.CONNECTION.CLIENT_ID);
		}
		else {
			throw new ContrabaseNetworkException(response.getErrorMessage());
		}
	}

	/**
	 * 연결을 끊어준다.
	 * 
	 * @throws ContrabaseNetworkException
	 */
	public void close() throws ContrabaseNetworkException {
		Message request = new Message();
		request.set(MSG.ACTION, MSG.REQUEST.CLOSE_CONNECTION);
		request.set(MSG.CONNECTION.CLIENT_ID, clientId);
		
		Message response = connection.send(request);
		if (response.isFailed()) {
			throw new ContrabaseNetworkException(response.getErrorMessage());
		}
		
		// 성공/실패 여부와 상관 없이, 연결은 닫아준다.
		connection.close();
	}

	/* (non-Javadoc)
	 * @see java.lang.Object#finalize()
	 */
	public void finalize() {
		try {
			close();
		} catch (ContrabaseNetworkException e) {
			e.printStackTrace();
		}
	}

	/**
	 * Monitoring Tool에서 사용하기 위한 Table List를 가져온다.
	 * @return
	 * @throws ContrabaseNetworkException
	 */
	@SuppressWarnings("unchecked")
	public List<String> getTableList() throws ContrabaseNetworkException {
		Message request = createActionMessage(MSG.REQUEST.GET_TABLE_LIST, "");
		Message response = connection.send(request);

		List<String> tableList = (List<String>) response.get(MSG.RESPONSE.TABLE_LIST);

		return tableList;
	}
	
	@SuppressWarnings("unchecked")
	public List<String> getRepositoryList() throws ContrabaseNetworkException{
		Message request = createActionMessage(MSG.REQUEST.GET_REPOSITORY_LIST, "");
		Message response = connection.send(request);
		List<String> repositoryList = (List<String>) response.get(MSG.RESPONSE.REPOSITORY_LIST);
		
		return repositoryList; 
	}
	
	@SuppressWarnings("unchecked")
	public Map<String, Long> getDataByAction(String action) throws ContrabaseNetworkException {
		Map<String, Long> result = null;
		Message request = new Message();
		request.set(MSG.CONNECTION.CLIENT_ID, clientId);
		
		if (action.equals("Size"))
			request.set(MSG.ACTION, MSG.REQUEST.GET_REPOSITORY_SIZE);
		else if (action.equals("Memory"))
			request.set(MSG.ACTION, MSG.REQUEST.GET_REPOSITORY_MEMORY);
		
		request.set(MSG.REQUEST.USER_ID, userId);

		Message response = connection.send(request);
		result = (Map<String, Long>)response.get(MSG.RESPONSE.ACTION_DATA);
		
		return result;
	}

	/**
	 * 테이블을 가져오거나, 없으면 생성한다.
	 * 
	 * @param tableName
	 * @return
	 * @throws ContrabaseNetworkException
	 */
	public Table getOrCreateTable(String tableName) throws ContrabaseNetworkException {
		Table table = getTable(tableName);
		return (table != null) ? table : createTable(tableName);
	}
	
	/**
	 * 새 테이블을 만든다.
	 * 
	 * @param tableName
	 * @return
	 * @throws ContrabaseNetworkException 
	 */
	public Table createTable(String tableName) throws ContrabaseNetworkException {
		Message request = createActionMessage(MSG.REQUEST.CREATE_TABLE, tableName);
		Message response = connection.send(request);
		
		if (response.isFailed())
			return null;
		
		return new Table(connection, clientId, userId, tableName);
	}

	/**
	 * 기존의 테이블을 가져온다. 마스터에 해당 테이블이 있는지 확인한다.
	 * 
	 * @param tableName
	 * @return
	 * @throws ContrabaseNetworkException 
	 */
	public Table getTable(String tableName) throws ContrabaseNetworkException {
		Message request = createActionMessage(MSG.REQUEST.GET_TABLE, tableName);
		Message response = connection.send(request);
		
		if (response.isFailed())
			return null;
		
		return new Table(connection, clientId, userId, tableName);
	}

	/**
	 * 테이블을 지운다.
	 * @param tableName
	 * @return
	 * @throws ContrabaseNetworkException 
	 */
	public void deleteTable(String tableName) throws ContrabaseNetworkException {
		Message request = createActionMessage(MSG.REQUEST.DELETE_TABLE, tableName);
		Message response = connection.send(request);
		
		if (response.isFailed())
			throw new ContrabaseNetworkException(request.getErrorMessage());
	}

	/**
	 * 테이블 조작하는 메시지를 생성한다.
	 * 
	 * @param tableName
	 * @return 생성된 메시지의 인스턴스를 넘겨준다.
	 */
	private Message createActionMessage(String action, String tableName) {
		Message request = new Message();
		request.set(MSG.CONNECTION.CLIENT_ID, clientId);
		request.set(MSG.ACTION, action);
		request.set(MSG.REQUEST.USER_ID, userId);
		request.set(MSG.REQUEST.TABLE_NAME, tableName);
		return request;
	}
	
	public void loadBalanceAllTables() throws ContrabaseNetworkException {
		Message request = new Message();
		request.set(MSG.ACTION, MSG.REQUEST.LOAD_BALANCE_ALL_TABLES);
		Message response = connection.send(request);
		
		if (response.isFailed())
			throw new RuntimeException(response.getErrorMessage());
	}
}
