package contrabase.api;

import java.io.IOException;
import java.util.ArrayList;
import java.util.HashMap;
import java.util.List;
import java.util.NoSuchElementException;
import java.util.Map.Entry;
import contrabase.ObjectContainer;
import contrabase.network.ClientSide;
import contrabase.network.ContrabaseNetworkException;
import contrabase.network.MSG;
import contrabase.network.MapReduce;
import contrabase.network.Message;
import contrabase.util.Enumerator;

public class ContrabaseResults {
	private MapReduce mapReduce;
	private HashMap<String, RepositoryCache> caches;
	private int maxCacheSize = 50;
	private List<RepositoryInfo> repositories;

	private static class RepositoryCache {
		private List<ObjectContainer> cache = new ArrayList<ObjectContainer>();
		private RepositoryInfo info;
		private boolean ended = false;
		
		public RepositoryCache(RepositoryInfo info) {
			this.info = info;
		}
		
		public RepositoryInfo getRepositoryInfo() {
			return info;
		}
		
		public List<ObjectContainer> getCache() {
			return cache;
		}

		public void addAll(List<ObjectContainer> objects) {
			cache.addAll(objects);
		}

		public boolean isEmpty() {
			return cache.isEmpty();
		}

		public ObjectContainer get(int index) {
			return cache.get(index);
		}

		public ObjectContainer pop() {
			ObjectContainer result = cache.get(0);
			cache.remove(0);
			return result;
		}

		public void setEnd() {
			ended = true;
		}
		
		public boolean isEnded() {
			return ended;
		}
	}
	
	
	/**
	 * 생성자 (최대 캐쉬 사이즈는 기본 값으로 설정된다.)
	 * 
	 * @throws ContrabaseNetworkException
	 */
	public ContrabaseResults(List<RepositoryInfo> repositories) throws ContrabaseNetworkException {
		initialize(repositories);
	}

	/**
	 * 생성자 (최대 캐쉬 사이즈를 지정한다.) 최대 캐쉬 사이즈가 0보다 작거나 같으면, 기본값으로 된다.
	 * 
	 * @throws ContrabaseNetworkException
	 */
	public ContrabaseResults(List<RepositoryInfo> repositories, int maxCacheSize) throws ContrabaseNetworkException {
		
		if (maxCacheSize > 0)
			this.maxCacheSize = maxCacheSize;	
		
		initialize(repositories);
	}

	/**
	 * 초기화 함수. (생성자에서 사용된다.)
	 * 
	 * @param messengers
	 * @throws ContrabaseNetworkException
	 * @throws InterruptedException
	 */
	final private void initialize(List<RepositoryInfo> repositories) throws ContrabaseNetworkException {
		this.repositories = repositories;
		
		// 캐쉬를 초기화 한다.
		caches = new HashMap<String, RepositoryCache>();
		for (RepositoryInfo rep : repositories)
			caches.put(rep.getRid(), new RepositoryCache(rep));
		
		// 리포지토리들과 연결을 한다.
		List<ClientSide> connections = new ArrayList<ClientSide>();
		
		for (RepositoryInfo rep : repositories)
			connections.add(new ClientSide(rep.getIp(), rep.getPort()));
		
		// 연결 정보로 MapReduce를 설정한다.
		mapReduce = new MapReduce(connections.toArray(new ClientSide[0]));
	}

	/**
	 * 캐쉬를 리필한다.
	 * @return 한 개 이상 리필 되었는지?
	 */
	private boolean refillCache(){
		//System.out.print("refillCache() : repos = ");
		
		// 요청을 만든다.
		Message request = new Message();
		request.set(MSG.ACTION, MSG.API_TO_REPOSITORY.GET_DATA);
		request.set(MSG.API_TO_REPOSITORY.COUNT, maxCacheSize);
		
		// 리포지토리 개수를 가져와서,
		int size = caches.size();
		
		//System.out.println("" + size);
		
		// 요청 대상들을 선정한다.
		boolean[] requestable = generateRequestable(size);
		
		// true 로 표시한 상대들에게 요청을 보낸다.
		Message[] responses = sendRequestsToRepositories(request, requestable);
		
		//System.out.println("refillCache() : recv");
		
		// 응답을 처리한다. : 데이터를 가져와서, 해당 캐쉬에 넣어준다.
		for (int i = 0; i < responses.length; i++) {
			
			// 하나도 안 남아서, 요청을 보낸 놈에 대해서만 처리 한다.
			if (requestable[i]) {
				String rid = repositories.get(i).getRid();
				
				@SuppressWarnings("unchecked")
				List<ObjectContainer> objects = (List<ObjectContainer>) responses[i].get(MSG.REPOSITORY_TO_API.DATA);
				
				//System.out.println("  @ " + objects.size());
				
				
				// 해당 repository 는 더 이상 가져올 값이 없다.
				if (objects.size() == 0) {
					caches.get(rid).setEnd();
					
					//System.out.println("  * end repo : " + rid);
				}
				
				// 데이터가 있는 경우.
				else {
					caches.get(rid).addAll(objects);
					
					//System.out.println("  * appended : " + rid);
				}
			}
		}
		
		//System.out.println("    refillCache's updated = " + updated);
		
		// 끝나지 않았으면, true.
		return isEnd() == false;
	}

	/**
	 * 전송 요청을 할 대상들을 true로 표시한다.
	 * @param size
	 * @return
	 */
	private boolean[] generateRequestable(int size) {
		boolean[] requestable = new boolean[size];
		
		for (Enumerator.Entry<RepositoryCache> entry : new Enumerator<RepositoryCache>(caches.values()))
			requestable[entry.getIndex()] = entry.getValue().isEmpty();
		
		return requestable;
	}
	
	/**
	 * 리포지토리들에 요청을 보낸다.
	 * @param request
	 * @param responses
	 * @param requestable
	 * @param size
	 * @return
	 */
	private Message[] sendRequestsToRepositories(Message request, boolean[] requestable) {
		int size = requestable.length;
		Message[] requests = new Message[size];
		
		try {
			for (int i = 0; i < size; i++) {
				// 메시지를 통째로 Deep-Copy 한 후에, 각각의 메시지에 채널을 입력 해준다.
				requests[i] = request.cloneMessage();
				requests[i].set(MSG.API_TO_REPOSITORY.CHANNEL, repositories.get(i).getChannel());
			}
			
			return mapReduce.map(requests, requestable);
		} catch (InterruptedException e) {
			e.printStackTrace();
			return null;
		}
	}
	
	/**
	 * 끝났는지 확인한다.
	 * @return
	 */
	private boolean isEnd() {
		//System.out.print("    " + caches.size() + " in isEnd() => ");
		
		boolean end = true;
		for (RepositoryCache aCache : caches.values()) {
			if (aCache.isEmpty() == false)
				end = false;
		}
		
		//System.out.println(end);
		return end;
	}

	/**
	 * 다음 노드가 있는지 확인한다.
	 * 
	 * @return 다음 노드가 있으면 true, 없으면 false
	 */
	public boolean hasNext() {
		// ended 를 확인해서, true 이면, false 를 리턴하고,
		// (ended 가 false 인데, 캐쉬가 0이면, 새로 요청한 후에..)
		// 캐쉬가 남아있으면, true를 리턴한다.
		
		// 끝났거나, 리포지토리 자체가 하나도 없으면, 종료한다.
		//if (ended || caches.isEmpty()) {
		if (caches.isEmpty()) {
			//System.out.println("hasNext is Ended!!");
			return false;
		}
		
		else {
			// O(2N)인데, 나중에 O(N)으로 변경하자.
			if (hasEmptyCache()) {
				//System.out.println("    hasEmptyCache() = true");
				boolean ret = refillCache();
				//System.out.println("    refillCache() = " + ret);
				return ret;
			}
			else {
				//System.out.println("    hasEmptyCache() = false");
				return true;
			}
		}
	}

	/**
	 * 캐쉬중에 비어있는 리스트가 하나라도 있는지 체크한다.
	 * @return 빈 거 있는가?
	 */
	private boolean hasEmptyCache() {
//		System.out.println("    in hasEmptyCache(), caches.size() = " + caches.size());
//		System.out.print("    ");
//		for (RepositoryCache aCache : caches.values())
//			System.out.print(aCache.getCache().size() + ", ");
//		System.out.println("");
		
		for (RepositoryCache aCache : caches.values()) {
			if (aCache.isEmpty())
				return true;
		}
		return false; 
	}

	/**
	 * 다음 노드를 ObjectContainer 에 담아서 가져온다.
	 * @return
	 * @throws ClassNotFoundException
	 * @throws IOException
	 */
	public ObjectContainer nextObjectContainer() throws ClassNotFoundException {
		// 캐쉬에 남는대로, 몇개를 요청할지 결정해서,
		// messenger 를 이용해서 메시지를 보내고,
		// 그 결과를 리스트에 차곡차곡 담는다.
		
		if (hasNext() == false)
			throw new NoSuchElementException();
		
		// Repository 에 직접 접속해서 데이터를 가져와야 하는데,
		// 외부 정렬을 해야한다.
		return getOneObject();
	}

	/**
	 * 다음 노드를 객체로 가져온다.
	 * 
	 * @return
	 * @throws ClassNotFoundException 
	 * @throws IOException 
	 */
	public Object next() throws ClassNotFoundException, IOException {
		return nextObjectContainer().toObject();
	}

	/**
	 * 외부 정렬을 해서, 객체 하나를 가져온다.
	 * @return
	 */
	private ObjectContainer getOneObject() {
		// XXX: 일단은, ID로 외부 정렬 한다. 다음 버전에서 order by 를 적용한다.
		
		// 각 리스트의 첫번째 ObjectConainer 의 key가 가장 작은 Repository의 Rid를 구한다.
		String foundRid = null;
		for (Entry<String, RepositoryCache> entry : caches.entrySet()) {
			
			// 끝나지 않은 항목들 중에서만..
			if (entry.getValue().isEnded() == false) {
				if (foundRid == null)
					foundRid = entry.getKey();
				
				String minKey = caches.get(foundRid).get(0).getKey();
				String newKey = entry.getValue().get(0).getKey();
				
				// newKey 가 더 작다!
				if (minKey.compareTo(newKey) > 0)
					foundRid = entry.getKey();
			}
		}
		
		if (foundRid != null) {
			// 찾아낸 항목을 가져오고, 지워준 후에,
			ObjectContainer result = caches.get(foundRid).pop();
			
			// 가져온 값을 리턴한다.
			return result;
		}
		
		throw new NoSuchElementException("Never reach here.");
	}
}
