package contrabase.api;

import java.io.IOException;
import java.io.Serializable;
import java.util.List;

import contrabase.ContrabaseException;
import contrabase.ObjectContainer;
import contrabase.condition.Condition;
import contrabase.network.ClientSide;
import contrabase.network.ContrabaseNetworkException;
import contrabase.network.MSG;
import contrabase.network.Message;
import contrabase.api.ContrabaseResults;
import static contrabase.condition.ConditionFactory.*;

public class Table {
	private ClientSide connection;
	private long clientId;
	private String userId;
	private String tableName;

	/**
	 * Table 의 생성자. default visibility 를 가진다.
	 * 유저는,
	 * {@link ContrabaseConnector#createTable(String)} 이나,
	 * {@link ContrabaseConnector#getTable(String)} 로 이 객체를 받는다.
	 * {@link ContrabaseConnector#close()} 이나,
	 * @param userId 
	 * @param tableName 
	 */
	Table(ClientSide connection, long clientId, String userId, String tableName) {
		this.connection = connection;
		this.clientId = clientId;
		this.userId = userId;
		this.tableName = tableName;
	}

	/////////////////////////////////////////////////////////////////
	// SELECT

	public Object selectByKey(String key) throws ContrabaseNetworkException, ContrabaseException {
		Condition cond = TERM("<key>", "==", key);
		Object value = null;
		
		ContrabaseResults results = selectByCondition(cond);
		if (results.hasNext()) {
			try {
				value = results.next();
			} catch (ClassNotFoundException e) {
				e.printStackTrace();
				throw new ContrabaseException(e);
			} catch (IOException e) {
				e.printStackTrace();
				throw new ContrabaseException(e);
			}
		}
		
		return value;
	}

	public ContrabaseResults selectByCondition(Condition condition) throws ContrabaseNetworkException {
		Message request = createRequest(MSG.REQUEST.SELECT);
		request.set(MSG.REQUEST.CONDITION, condition);
		
		Message response = connection.send(request);
		if (response.isFailed())
			throw new ContrabaseNetworkException(response.getErrorMessage());
		
		// 리포지토리가 여러개임. IP도 포트도 여러개임.
		@SuppressWarnings("unchecked")
		List<RepositoryInfo> repositories = (List<RepositoryInfo>) response.get(MSG.RESPONSE.REPOSITORIES);
		
		// 결과 이터레이터를 돌려준다.
		ContrabaseResults resultsIterator = new ContrabaseResults(repositories);
		
		return resultsIterator;
	}

	public ContrabaseResults selectTableContent() throws ContrabaseException {
		Message request = createRequest(MSG.REQUEST.SELECT);
		Condition condition = TERM("<key>", "==", "*");		// 모든 key 를 가져온다.
		request.set(MSG.REQUEST.CONDITION, condition);

		Message response = connection.send(request);
		if (response.isFailed())
			throw new ContrabaseNetworkException(response.getErrorMessage());
		
		// 리포지토리가 여러개임. IP도 포트도 여러개임.
		@SuppressWarnings("unchecked")
		List<RepositoryInfo> repositories = (List<RepositoryInfo>) response.get(MSG.RESPONSE.REPOSITORIES);
		
		// 결과 이터레이터를 돌려준다.
		ContrabaseResults resultsIterator = new ContrabaseResults(repositories);
		
		return resultsIterator;		
	}

	/////////////////////////////////////////////////////////////////
	// INSERT

	public boolean insert(String key, Serializable value) throws ContrabaseException {
		ObjectContainer oc = null;
		
		try {
			oc = new ObjectContainer(key, value);
		} catch (IOException e) {
			throw new ContrabaseException(e);
		} catch (ClassNotFoundException e) {
			throw new ContrabaseException(e);
		}
		
		Message request = createRequest(MSG.REQUEST.INSERT);
		request.set(MSG.REQUEST.INSERT_DATA, oc);
		
		Message response = connection.send(request);
		
		return response.isSuccessed();
	}


	/////////////////////////////////////////////////////////////////
	// UPDATE

	public int updateByKey(String key, List<UpdateValue> updateValues)
	throws ContrabaseNetworkException, ContrabaseException {
		Condition cond = TERM("<key>", "==", key);
		return updateByCondition(cond, updateValues);
	}

	public int updateByCondition(Condition condition, List<UpdateValue> updateValues)
	throws ContrabaseNetworkException, ContrabaseException {
		Message request = createRequest(MSG.REQUEST.UPDATE);
		request.set(MSG.REQUEST.CONDITION, condition);
		request.set(MSG.REQUEST.UPDATE_VALUES, updateValues);
		
		Message response = connection.send(request);
		
		if (response.isSuccessed())
			return (Integer)response.get(MSG.RESPONSE.COUNT);
		else
			throw new ContrabaseException(response.getErrorMessage());
	}


	/////////////////////////////////////////////////////////////////
	// DELETE

	public int deleteByKey(String key)
	throws ContrabaseNetworkException, ContrabaseException {
		Condition cond = TERM("<key>", "==", key);
		return deleteByCondition(cond);
	}

	public int deleteByCondition(Condition condition) throws ContrabaseNetworkException, ContrabaseException {
		Message request = createRequest(MSG.REQUEST.DELETE);
		
		request.set(MSG.REQUEST.CONDITION, condition);
		Message response = connection.send(request);
		
		if (response.isSuccessed())
			return (Integer)response.get(MSG.RESPONSE.COUNT);
		else
			throw new ContrabaseException(response.getErrorMessage());
	}


	/////////////////////////////////////////////////////////////////
	// LOAD BALANCING

	/**
	 * 테이블 한 개에 대해서, 로드 밸런싱 요청을 날린다.
	 * @return 로드 밸런싱을 했는지? 안했는지?
	 * @throws ContrabaseNetworkException
	 */
	public boolean loadBalance() throws ContrabaseNetworkException {
		Message request = createRequest(MSG.REQUEST.LOAD_BALANCE_TABLE);
		Message response = connection.send(request);
		return response.isSuccessed();
	}

	/////////////////////////////////////////////////////////////////
	/////////////////////////////////////////////////////////////////

	private Message createRequest(String action) {
		Message request = new Message();
		request.set(MSG.CONNECTION.CLIENT_ID, clientId);
		request.set(MSG.REQUEST.USER_ID, userId);
		request.set(MSG.REQUEST.TABLE_NAME, tableName);
		request.set(MSG.ACTION, action);
		return request;
	}
}
