package contrabase.master;

import java.util.ArrayList;
import java.util.Collections;
import java.util.Comparator;
import java.util.HashMap;
import java.util.List;
import java.util.Map;

import contrabase.network.ClientSide;
import contrabase.network.MSG;
import contrabase.network.MapReduce;
import contrabase.network.Message;
import contrabase.repository.datastore.ContrabaseDatastoreException;
import contrabase.util.Enumerator;
import contrabase.util.Pair;

/**
 * @author jong10
 * 로드 밸런싱을 하는 클래스
 */
public class LoadBalancer {
	private RepositoriesManager manager;
	private String dataDir;
	
	private List<Pair<String, String>> totalKeys;			// Pair< key, RID >
	private Map<String, Pair<String, Integer>> addresses;	// Map< RID, Pair< IP, Port > >
	
	public LoadBalancer(RepositoriesManager manager, String dataDir) {
		this.manager = manager;
		this.dataDir = dataDir;
	}
	
	/**
	 * 테이블 한 개 이상이, 로드 밸런싱이 필요한지 체크한다.
	 * 실행하는 동안에는, Master의 DB에 락을 걸지는 않아도 될듯?? 읽기만 필요하니깐.
	 * @throws ContrabaseDatastoreException 
	 */
	public List<Pair<String, String>> getNeedsLoadBalancingTables() throws ContrabaseDatastoreException {
		MasterSectionDAO section = new MasterSectionDAO(dataDir, null, null);
		List<Pair<String, String>> list = section.getAllTableList();
		section.close();
		return list;
	}
	
	/**
	 * 해당 이름의 테이블을 로드 밸런스 한다.
	 * @param tableName
	 * @throws ContrabaseDatastoreException 
	 * @throws InterruptedException 
	 */
	public void loadBalanceTable(String userId, String tableName) throws ContrabaseDatastoreException, InterruptedException {
		
		// Master -> Repositories : 정렬된 키 목록 내놔
		collectKeysFromAllRepositories(userId, tableName);
		
		// rid 목록을 가져온다. 오름차순으로.
		MasterSectionDAO section = new MasterSectionDAO(dataDir, userId, tableName);
		List<String> rids = section.getRepositoryList();
		
		// diff 목록을 만든다. : Map< 소유자RID, ArrayList<Pair< KEY, 받을놈RID >>>
		Map<String, ArrayList<Pair<String, String>>> diffs = generateDiffs(rids);
		
//		// diffs 내용 확인. (성공인듯?)
//		System.out.println("oldRID\tKEY\tnewRID");
//		for (String rid : diffs.keySet()) {
//			ArrayList<Pair<String, String>> values = diffs.get(rid);
//			for (Pair<String, String> p : values) {
//				System.out.println(rid + "\t" + p.getLeft() + "\t" + p.getRight());
//			}
//		}
		
		// 소유자 RID에 대한, IP+port 정보와 diff 목록을 만들어서, MapReduce 때린다. (나머지는 R들에게 맡긴다.)
		processLoadBalancing(diffs, userId, tableName);
		
		// 이 테이블의, 모든 RID에 대한 endKey 를 갱신한다.
		updateEndKeys(diffs, section);
	}

	/**
	 * 이 테이블의, 모든 RID에 대한 endKey 를 갱신한다.
	 * @param diffs
	 * @param section 
	 */
	private void updateEndKeys(Map<String, ArrayList<Pair<String, String>>> diffs, MasterSectionDAO section) {
		Map<String, String> updateList = new HashMap<String, String>();
		// < RID  , endKey>
		
		for (Pair<String, String> entry : totalKeys) {
			String key = entry.getLeft();
			String rid = entry.getRight();
			
			// 없으면, 이 key 를 endKey 로 설정.
			if (updateList.containsKey(rid) == false) {
				updateList.put(rid, key);
			}
			
			// key 가 저장된 endKey 보다 더 작으면 업데이트.
			else if (updateList.get(rid).compareTo(key) < 0) {
				updateList.put(rid, key);
			}
		}
		
		// endKey 를 재설정 한다.
		for (Map.Entry<String, String> entry : updateList.entrySet()) {
			String rid = entry.getKey();
			String endKey = entry.getValue();
			section.updateInfo(rid, endKey);
		}
	}

	/**
	 * 로드 밸런싱을 진행 한다.
	 * @param diffs
	 * @param userId 
	 * @param tableName 
	 * @throws InterruptedException
	 */
	private void processLoadBalancing(Map<String, ArrayList<Pair<String, String>>> diffs, String userId, String tableName) throws InterruptedException {
		int size = diffs.size();
		
		Message[] requests = new Message[size];
		ClientSide[] connections = new ClientSide[size];
		boolean[] requestables = new boolean[size];
		
		// requestables 는 모두 true 로 채운다.
		for (int i = 0; i < requestables.length; i++)
			requestables[i] = true;
		
		// connections 과 requests 를 채운다.
		for (Enumerator.Entry<String> repo : new Enumerator<String>(diffs.keySet())) {
			int index = repo.getIndex();
			String rid = repo.getValue();
			ArrayList<Pair<String, String>> aDiffList = diffs.get(rid);
			
			// Repository 의 RequestHandler 와 연결된 ClientSide 를 설정한다.
			connections[index] = manager.getRepositoryEntry(rid).getClientSide();
			
			// 액션을 지정해준다.
			requests[index] = new Message();
			requests[index].set(MSG.ACTION, MSG.REQUEST.LOAD_BALANCE_TABLE);
			requests[index].set(MSG.REQUEST.USER_ID, userId);
			requests[index].set(MSG.REQUEST.TABLE_NAME, tableName);
			
			// 이 Repository 에서, 다른 Repository 로 보내줄 데이터 목록.
			requests[index].set(MSG.MANAGEMENT.LOAD_BALANCE_DATA, aDiffList);
			
			// 다른 Repository 들에 대한 접속 정보 (IP, Port)
			requests[index].set(MSG.MANAGEMENT.REPOSITORY_ADDRESSES, addresses);
		}
		
		// 날려 날려
		MapReduce mapReduce = new MapReduce(connections);
		Message[] responses = mapReduce.map(requests, requestables);
		
		// 결과에 에러가 있는지 확인한다.
		for (int i = 0; i < responses.length; i++) {
			if (responses[i].isFailed())
				throw new RuntimeException("responses[" + i + "] is failed!");
		}
	}
	
	/**
	 * 변한 부분의 목록을 만든다.
	 * Map< 소유자RID, Pair< KEY, 받을놈RID >>
	 * @return
	 */
	private Map<String, ArrayList<Pair<String, String>>> generateDiffs(List<String> rids) {
		Map<String, ArrayList<Pair<String, String>>> diffs = new HashMap<String, ArrayList<Pair<String, String>>>();
		
		// 리포지토리 한개가 얼만큼의 데이터를 가져야 하는지..
		final int INTERVAL = countInterval(totalKeys.size(), manager.getSize());
		
		// 해당 항목을 diffs에 추가 해준다.
		for (int i = 0; i < totalKeys.size(); i++) {
			String key = totalKeys.get(i).getLeft();
			String oldRid = totalKeys.get(i).getRight();
			
			// 목록을 가져오거나, 없으면 생성하고..
			ArrayList<Pair<String, String>> list = getOrPut(diffs, oldRid);
			
			// 받을놈 RID 를 구하고..
			String newRid = rids.get(i / INTERVAL);
			
			// oldRid와 newRid가 다른 놈들만 목록에 추가한다.
			if (oldRid.equals(newRid) == false)
				list.add(new Pair<String, String>(key, newRid));
		}
		
		return diffs;
	}

	private int countInterval(int total, int size) {
		if (total % size != 0)
			total += size - (total % size);
		return total / size;
	}

	private ArrayList<Pair<String, String>> getOrPut(Map<String, ArrayList<Pair<String, String>>> diffs, String rid) {
		if (diffs.containsKey(rid)) {
			return diffs.get(rid);
		}
		else {
			ArrayList<Pair<String, String>> newList = new ArrayList<Pair<String, String>>();
			diffs.put(rid, newList);
			return newList;
		}
	}

	/**
	 * 모든 리포지토리가 가지고 있는 키 목록을 가져와서 합병 및 정렬한다.
	 * @return
	 */
	@SuppressWarnings("unchecked")
	private void collectKeysFromAllRepositories(String userId, String tableName) {
		totalKeys = new ArrayList<Pair<String, String>>();
		addresses = new HashMap<String, Pair<String, Integer>>();
		
		// Repository 들의 Request Handler 들과 연결을 한다.
		MapReduce mapReduce = new MapReduce(manager.getConnections());
		
		
		// 메시지를 뿌리고, 결과를 취합한다.
		Message msg = new Message();
		msg.set(MSG.ACTION, MSG.MANAGEMENT.GET_KEYS);
		msg.set(MSG.REQUEST.USER_ID, userId);
		msg.set(MSG.REQUEST.TABLE_NAME, tableName);
		
		Message[] responses = null;
		
		try {
			responses = mapReduce.map(msg);
		} catch (InterruptedException e) {
			e.printStackTrace();
		}
		
		if (responses == null)
			return;
		
		// 성공한 요소들을 totalKeys 와 addresses 에 반영한다.
		for (Message res : responses) {
			if (res != null && res.isSuccessed()) {
				String rid = (String)res.get(MSG.RESPONSE.RID);
				List<String> keys = (List<String>)res.get(MSG.RESPONSE.KEY_LIST);
				
				RepositoryEntry entry = manager.getRepositoryEntry(rid);
				String ip = entry.getDataIp();
				int port = entry.getDataPort();
				
				// IP + 포트 갱신
				if (ip != null && port != 0)
					addresses.put(rid, new Pair<String, Integer>(ip, port));
				
				// totalKeys 갱신
				if (keys != null)
					for (String aKey : keys)
						totalKeys.add(new Pair<String, String>(aKey, rid));
			}
		}
		
		// totalKeys 를 첫번째 칸(key) 기준으로 정렬 한다.
		Collections.sort(totalKeys, new Comparator<Pair<String, String>>() {
			@Override
			public int compare(Pair<String, String> o1, Pair<String, String> o2) {
				return o1.getLeft().compareTo(o2.getLeft());
			}
		});
	}

	/**
	 * 모든 유저의 모든 테이블을 로드 밸런스 한다.
	 * @throws InterruptedException 
	 * @throws ContrabaseDatastoreException 
	 */
	public void loadBalanceAllTables() throws InterruptedException, ContrabaseDatastoreException {
		// 전체 락을 잠그고,
		manager.startLoadBalancing();
		
		// 로드 밸런싱이 필요한 모든 테이블을 구하고,
		List<Pair<String, String>> tables = getNeedsLoadBalancingTables();
		
		// 해당 테이블들을 로드 밸런싱 한다.
		for (Pair<String, String> table : tables) {
			String userId = table.getLeft();
			String tableName = table.getRight();
			
			loadBalanceTable(userId, tableName);
		}
		
		// 전체 락을 푼다.
		manager.endLoadBalancing();
	}
}
