package contrabase.master;

import java.sql.Connection;
import java.sql.DriverManager;
import java.sql.PreparedStatement;
import java.sql.ResultSet;
import java.sql.SQLException;
import java.util.ArrayList;
import java.util.List;

import contrabase.repository.datastore.ContrabaseDatastoreException;
import contrabase.util.Pair;

/**
 * MasterTable에서 생성한 Table 을 관리하는 클래스
 * 
 * masterSection(
 * 		userId TEXT NOT NULL,
 * 		tableName TEXT,
 * 		repositoryId TEXT,
 * 		endKey TEXT,
 * 		primary key(userId, tableName, repositoryId))
 * 
 */
public class MasterSectionDAO {

	private Connection conn = null;
	private String userId;
	private String tableName;
	private String dataDir;

	/**
	 * Monitoring Tool에서 사용하는 생성자.
	 * 
	 * @param dataDir
	 *            sqlite가 사용하는 .db가 위치하는 Directory Name
	 * @throws ContrabaseDatastoreException
	 */
	public MasterSectionDAO(String dataDir, String userId) throws ContrabaseDatastoreException {
		this.userId = userId;
		this.dataDir = dataDir;
		this.tableName = "";
		open();
	}

	/**
	 * @param dataDir
	 *            sqlite가 사용하는 .db가 위치하는 Directory Name
	 * @param userId
	 *            User ID
	 * @param tableName
	 *            Talbe Name
	 * @throws ContrabaseDatastoreException
	 */
	public MasterSectionDAO(String dataDir, String userId, String tableName) throws ContrabaseDatastoreException {
		this.userId = userId;
		this.tableName = tableName;
		this.dataDir = dataDir;
		open();
	}

	/**
	 * master.db 의 Connection 을 얻는 method
	 */
	private void open() throws ContrabaseDatastoreException {
		String url = "jdbc:sqlite:" + dataDir + "/master.db";
		try {
			Class.forName("org.sqlite.JDBC");
			conn = DriverManager.getConnection(url);
		} catch (ClassNotFoundException e) {
			throw new ContrabaseDatastoreException(e);
		} catch (SQLException e) {
			throw new ContrabaseDatastoreException(e);
		}
	}

	@Override
	public void finalize() {
		try {
			close();
		} catch (ContrabaseDatastoreException e) {
			e.printStackTrace();
		}
	}

	/**
	 * master.db 의 Connection 을 닫는 method
	 */
	public void close() throws ContrabaseDatastoreException {
		try {
			if (conn != null) {
				conn.close();
				conn = null;
			}
		} catch (SQLException e) {
			throw new ContrabaseDatastoreException(e);
		}
	}

	/**
	 * masterSection Table 에서 해당 유저의 Table List를 가져오는 메소드
	 * @return
	 */
	public List<String> getTableList() {
		List<String> tableList = new ArrayList<String>();
		PreparedStatement pstat = null;

		String sql = "select distinct tableName from masterSection where userId = ?";

		try {
			pstat = conn.prepareStatement(sql);
			pstat.setString(1, userId);
			
			ResultSet rs = pstat.executeQuery();
			while (rs.next()) {
				tableList.add(rs.getString(1));
			}
		} catch (SQLException e) {
			e.printStackTrace();
			return tableList;
		}finally {
			try {
				pstat.close();
			} catch (SQLException e) {
				e.printStackTrace();
			}
		}
		return tableList;
	}

	/**
	 * masterSection Table 에서 해당 유저의 repository List를 가져오는 메소드
	 * @return
	 */
	public List<String> getRepositoryList() {
		List<String> repositoryList = new ArrayList<String>();
		PreparedStatement pstat = null;
		
		String sql = "select distinct repositoryId, endKey from masterSection where userId=? order by endKey asc";
		
		try {
			pstat = conn.prepareStatement(sql);
			pstat.setString(1, userId);
			
			ResultSet rs = pstat.executeQuery();
			while (rs.next()) {
				repositoryList.add(rs.getString(1));
			}
		} catch (SQLException e) {
			e.printStackTrace();
			return repositoryList;
		} finally {
			try {
				pstat.close();
			} catch (SQLException e) {
				e.printStackTrace();
			}
		}
		
		return repositoryList; 
	}
	/**
	 * masterSection Table 에 repository 에 담을 테이블을 추가하는 메소드
	 * 
	 * @param rid
	 * @param endKey
	 */
	public boolean addInfo(String rid, String endKey) {
		PreparedStatement pstat = null;
		boolean result = true;
		try {
			pstat = conn.prepareStatement("insert into masterSection values(?, ?, ?, ?)");
			pstat.setString(1, userId);
			pstat.setString(2, tableName);
			pstat.setString(3, rid);
			pstat.setString(4, endKey);

			pstat.execute();

		} catch (SQLException e) {
			result = false;
		} finally {
			try {
				if (pstat != null)
					pstat.close();
			} catch (SQLException e) {
				e.printStackTrace();
			}
		}
		return result;
	}

	/**
	 * 해당 rid 에서 Key값을 삭제하는 메소드
	 * 
	 * @param rid
	 *            RepositoryID
	 * @return boolean
	 */
	public boolean deleteInfo(String rid) {
		boolean result = true;
		PreparedStatement pstat = null;
		try {
			String sql = "delete from masterSection where userId = ? AND tableName = ? AND repositoryId = ?";
			pstat = conn.prepareStatement(sql);
			pstat.setString(1, userId);
			pstat.setString(2, tableName);
			pstat.setString(3, rid);
			pstat.execute();
		} catch (SQLException e) {
			result = false;
		} finally {
			try {
				pstat.close();
			} catch (SQLException e) {
				e.printStackTrace();
			}
		}
		return result;
	}

	/**
	 * 해당 rid 에 대한, endKey 를 가져오는 메소드.
	 * 
	 * @param rid
	 *            RepositoryID
	 * @param endKey
	 *            Key 구분자
	 * @return boolean
	 */
	public boolean updateInfo(String rid, String endKey) {
		boolean result = true;
		PreparedStatement pstat = null;
		try {
			String sql = "update masterSection set endKey = ? where userId = ? AND tableName = ? AND repositoryId = ?";
			pstat = conn.prepareStatement(sql);

			pstat.setString(1, endKey);
			pstat.setString(2, userId);
			pstat.setString(3, tableName);
			pstat.setString(4, rid);

			pstat.execute();
		} catch (SQLException e) {
			result = false;
		} finally {
			try {
				pstat.close();
			} catch (SQLException e) {
				e.printStackTrace();
			}
		}
		return result;
	}

	/**
	 * 해당 rid 에 대한, endKey 를 가져오는 메소드.
	 * 
	 * @param rid RepositoryID
	 * @return EndKey 반환
	 */
	public String getEndKey(String rid) {
		String endKey = "";
		PreparedStatement pstat = null;
		ResultSet rs = null;

		try {
			String sql = "select endKey from masterSection where userId = ? AND tableName = ? AND repositoryId = ?";
			pstat = conn.prepareStatement(sql);

			pstat.setString(1, userId);
			pstat.setString(2, tableName);
			pstat.setString(3, rid);

			rs = pstat.executeQuery();

			if (rs.next())
				endKey = rs.getString("endKey");

		} catch (SQLException e) {
			e.printStackTrace();
		} finally {
			try {
				rs.close();
				pstat.close();
			} catch (SQLException e) {
				e.printStackTrace();
			}
		}
		return endKey;
	}

	/**
	 * 해당 유저이름의, 해당 테이블 이름으로 된 모든 행을 지운다. 실질적으로, 해당 테이블을 지우는 내용임 존재하지 않았으면,
	 * false 를 리턴
	 */
	public boolean deleteAll() {
		boolean result = true;
		PreparedStatement pstat = null;

		try {
			String sql = "delete from masterSection where userId = ? AND tableName = ?";
			pstat = conn.prepareStatement(sql);
			pstat.setString(1, userId);
			pstat.setString(2, tableName);
			pstat.execute();
		} catch (SQLException e) {
			result = false;
		} finally {
			try {
				pstat.close();
			} catch (SQLException e) {
				e.printStackTrace();
			}
		}
		return result;
	}

	/**
	 * 해당 <유저이름, 테이블이름> 의 항목이 존재하는지 확인한다.
	 * 
	 * @return 존재하면 true 존재하지 않으면 false를 리턴한다.
	 */
	public boolean isExist() {
		boolean result = false;
		PreparedStatement pstat = null;
		ResultSet rs = null;

		try {
			String sql = "select * from masterSection where userId = ? AND tableName = ?";
			pstat = conn.prepareStatement(sql);
			pstat.setString(1, userId);
			pstat.setString(2, tableName);
			rs = pstat.executeQuery();

			if (rs.next())
				result = true;

		} catch (SQLException e) {
			e.printStackTrace();
			result = false;
		} finally {
			try {
				pstat.close();
			} catch (SQLException e) {
				e.printStackTrace();
			}
		}
		return result;
	}

	/**
	 * 모든, <유저, 테이블이름> 목록을 가져온다.
	 * @return
	 */
	public List<Pair<String, String>> getAllTableList() {
		List<Pair<String, String>> allTables = new ArrayList<Pair<String, String>>();
		
		String sql = "select userId, tableName from masterSection";
		PreparedStatement pstat;
		try {
			pstat = conn.prepareStatement(sql);
			ResultSet rs = pstat.executeQuery();
			
			while (rs.next())
				allTables.add(new Pair<String, String>(rs.getString("userId"), rs.getString("tableName")));
		} catch (SQLException e) {
			e.printStackTrace();
		}
		
		return allTables;
	}
}
