package contrabase.master;

import contrabase.network.ClientSide;
import contrabase.network.ContrabaseNetworkException;
import contrabase.network.MSG;
import contrabase.network.Message;
import contrabase.network.MessageHandler;
import contrabase.repository.TabletEntries;

/**
 * @author jong10
 * Master 에서 Repository 의 요청을 처리하는 클래스
 */
public class RepositoryHandler implements MessageHandler {
	private int lastRepositoryNo = 0;
	private RepositoriesManager manager;
	private Object lastRepositoryNoLock = new Object[0];
	private MasterGUI masterGUI;

	public RepositoryHandler(RepositoriesManager manager) {
		this.manager = manager;
	}

	public RepositoryHandler(RepositoriesManager manager, MasterGUI masterGUI) {
		this.manager = manager;
		this.masterGUI = masterGUI;
	}

	/**
	 * 처음 리포지토리와 연결이 되면, 양방향 연결을 시도한다.
	 * MessageHandler 를 상속받은 클래스
	 * @param request  : 메시지를 요청하는 request Message 
	 * @param response : 메시지를 처리해서 결과값을 리턴하는 response Message
	 * 
	 * */
	@Override
	public void handle(Message request, Message response) {
		if (MSG.MANAGEMENT.CONNECTION_INIT.equals(request.get(MSG.ACTION))) {
			try {
				initializeBidirectional(request, response);
			} catch (ContrabaseNetworkException e) {
				// XXX: 여기서 네트웍 에러가 나면 어떻게 처리해야 하지?? 죽여야 하나??
				// 아니면, 해당 리포지토리만 연결을 끊어야 하나??
				e.printStackTrace();
				System.exit(1);
			}
		}
		
		// 그 이후에는, 일반적인 메시지 처리를 한다.
		else {
			handleMessage(request, response);
		}
	}

	/**
	 * 실제로, Master 에서 Repository 의 메시지를 처리하는 메서드.
	 * 양방향 연결이 연결된 후이다.
	 * 
	 * @param request
	 * @param response
	 */
	private void handleMessage(Message request, Message response) {
		// XXX: Master 에서 Repository 의 메시지를 처리하는 메서드. 현재 버전에서 사용하지 않음.
		// Master 가 Repository 의 요청을 받는 경우를 없애기로 했음.
	}

	/**
	 * Repository 에서 접속을 처음 받았을때, "양방향 연결"을 초기화 해주는 메소드.
	 * 
	 * @param request
	 * @param response
	 * @throws ContrabaseNetworkException 
	 */
	private void initializeBidirectional(Message request, Message response)	throws ContrabaseNetworkException {
		// request 에 담겨져 전달된 RID, IP, port 를 받아온다.
		String rid = (String)request.get(MSG.MANAGEMENT.REPOSITORY_RID);
		String ip = (String)request.get(MSG.MANAGEMENT.REPOSITORY_IP);
		int port = (Integer)request.get(MSG.MANAGEMENT.REPOSITORY_PORT);
		
		// dataHandler 의 IP와 Port 정보.
		String dataIP = (String)request.get(MSG.MANAGEMENT.REPOSITORY_DATA_IP);
		int dataPort = (Integer)request.get(MSG.MANAGEMENT.REPOSITORY_DATA_PORT);
		
		TabletEntries tablets = (TabletEntries) request.get(MSG.MANAGEMENT.TABLET_ENTRIES);
		
		ClientSide connection = new ClientSide(ip, port);
		RepositoryEntry entry = new RepositoryEntry(rid, connection, tablets, dataIP, dataPort);
		manager.setRepositoryEntry(entry);
		
		System.out.println("New repository connected : " + rid + " (tablet size = " + tablets.size() + ")");
		// Master GUI에 Log를 출력하는 부분.
		masterGUI.appendLog("New repository connected : " + rid + " (tablet size = " + tablets.size() + ")");
		
		// 동기화
		synchronized (lastRepositoryNoLock) {
			// repositoryNo 를 설정 해준다. 1부터 시작한다!
			response.set(MSG.MANAGEMENT.REPOSITORY_NO, ++lastRepositoryNo);
		}
		
		// Repository 에 응답을 보내준다.
		response.setSuccessed();
	}
}
