package contrabase.master;

import java.util.HashMap;
import java.util.Map;
import java.util.Set;
import java.util.TreeSet;

import contrabase.master.actionhandler.*;
import contrabase.network.*;

/**
 * Master 에서 API 의 요청을 처리하는 클래스
 */
public class ServiceHandler implements MessageHandler {
	private RepositoriesManager manager;
	private long lastClientId = 0;
	private Set<Long> connectedClients;
	private Map<String, ActionHandler> actionHandlers;
	private MasterGUI masterGUI;
	
	public ServiceHandler(RepositoriesManager manager, String dataDir) {
		this.manager = manager;
		connectedClients = new TreeSet<Long>();
		actionHandlers = new HashMap<String, ActionHandler>();
		initializeActionHandlers(dataDir);
	}
	
	public ServiceHandler(RepositoriesManager manager, String dataDir, MasterGUI masterGUI) {
		this.manager = manager;
		connectedClients = new TreeSet<Long>();
		actionHandlers = new HashMap<String, ActionHandler>();
		initializeActionHandlers(dataDir);
		this.masterGUI = masterGUI;
	}

	/**
	 * @param request  : 메시지를 요청하는 request Message 
	 * @param response : 메시지를 처리해서 결과값을 리턴하는 response Message
	 * 
	 * Action Field 를 확인하고 
	 * 초기 접속할 경우에는 ClientID를  생성해서 생성한 ClientID를 돌려준다.
	 * CClientID가 있으면 해당 request 에 대한 요청을 파악해서 처리한다.
	 */
	@Override
	public void handle(Message request, Message response) {

		// ACTION 필드를 확인한다.
		String action = (String)request.get(MSG.ACTION);
		if (action == null || action.equals("")) {
			response.setFailed("ACTION is not exist");
			// Master GUI에 Log를 출력하는 부분.
			masterGUI.appendLog(request.get(MSG.REQUEST.USER_ID)+ " : " + "ACTION is not exist");
			return;
		}
		
		// 전체 테이블 로드 밸런싱인 경우.
		else if (action.equals(MSG.REQUEST.LOAD_BALANCE_ALL_TABLES)) {
			ActionHandler handler = actionHandlers.get(action);
			if (handler != null)
				handler.action(request, response);
			else
				response.setFailed("UNDEFINED ACTION: " + action);
			
			return;
		}
		
		// 클라이언트 아이디가 없는 경우. (연결 요청인 경우).
		// 클라이언트 아이디를 생성해서, 응답으로 돌려준다.
		else if (request.get(MSG.CONNECTION.CLIENT_ID) == null) {
			if (action.equals(MSG.REQUEST.OPEN_CONNECTION)) {
				long clientId = nextClientId();
				connectedClients.add(clientId);
				response.set(MSG.CONNECTION.CLIENT_ID, clientId);
				response.setSuccessed();
			}
			else {
				response.setFailed("need OPEN_CONNECTION");
				// Master GUI에 Log를 출력하는 부분.
				masterGUI.appendLog("need OPEN_CONNECTION");
				return;
			}
		}
		
		// 그 이외의 요청. (클라이언트 아이디가 있는 경우)
		else {
			long clientId = (Long)request.get(MSG.CONNECTION.CLIENT_ID);
			
			// 해당 클라이언트 아이디가 아직 접속이 되어있는 상태로 기록되어 있는지 확인.
			if (connectedClients.contains(clientId)) {
				
				// 접속 종료 요청
				if (action.equals(MSG.REQUEST.CLOSE_CONNECTION)) {
					connectedClients.remove(clientId);
					response.setSuccessed();
				}
				
				// 나머지 요청을 처리한다.
				else {
					// 동기화 중이면, 끝날때까지 기다린다.
					try {
						manager.waitForLoadBalancing();
					} catch (InterruptedException e) {
						//e.printStackTrace();
						throw new RuntimeException(e);
					}
					
					// Master GUI에 Log를 출력하는 부분.
					masterGUI.appendLog(request.get(MSG.REQUEST.USER_ID)+ " : "+ action);
					
					// 실제 요청들을 처리한다.
					ActionHandler handler = actionHandlers.get(action);
					
					if (handler != null)
						handler.action(request, response);
					else
						response.setFailed("UNDEFINED ACTION: " + action);
				}
			}
			
			// 접속이 이미 끊긴 경우.
			else {
				response.setFailed("connection closed");
			}
		}
	}

	/**
	 * 핸들러들을 초기화 한다.
	 */
	private void initializeActionHandlers(String dataDir) {
		actionHandlers.put(MSG.REQUEST.GET_TABLE, new ActionGetTable(manager, dataDir));
		actionHandlers.put(MSG.REQUEST.GET_TABLE_LIST, new ActionGetTableList(manager, dataDir));
		actionHandlers.put(MSG.REQUEST.CREATE_TABLE, new ActionCreateTable(manager, dataDir));
		actionHandlers.put(MSG.REQUEST.DELETE_TABLE, new ActionDeleteTable(manager, dataDir));
		
		actionHandlers.put(MSG.REQUEST.GET_REPOSITORY_LIST, new ActionGetRepositoryList(manager, dataDir));
		actionHandlers.put(MSG.REQUEST.GET_REPOSITORY_SIZE, new ActionGetRepositorySize(manager));
		actionHandlers.put(MSG.REQUEST.GET_REPOSITORY_MEMORY, new ActionGetRepositoryMemory(manager));
		
		actionHandlers.put(MSG.REQUEST.LOAD_BALANCE_TABLE, new ActionLoadBalanceTable(manager, dataDir));
		actionHandlers.put(MSG.REQUEST.LOAD_BALANCE_ALL_TABLES, new ActionLoadBalanceAllTables(manager, dataDir));
		
		actionHandlers.put(MSG.REQUEST.INSERT, new ActionInsert(manager));
		actionHandlers.put(MSG.REQUEST.UPDATE, new ActionUpdate(manager));
		actionHandlers.put(MSG.REQUEST.SELECT, new ActionSelect(manager));
		actionHandlers.put(MSG.REQUEST.DELETE, new ActionDelete(manager));
	}

	/**
	 * 다음번의 클라이언트 접속에 대한 아이디를 가져온다. 1부터 시작한다.
	 * @return 다음에 사용될 클라이언트 접속에 대한 아이디.
	 */
	private synchronized long nextClientId() {
		// XXX: 일단은, 오버플로우에 대한 고려는 하지 않음.
		return ++lastClientId;
	}
}
