package contrabase.network;

import contrabase.master.Lock;

/**
 *  API 에서 Master 에게 Select 요청을 했을때
 *  Client 가 Master 에게 해당 repositoryIP와 Port 를 얻어와서 접속해서 데이터를 가지고 오는 클래스
 */
public class MapReduce {
	private ClientSide[] connections;
	private int remainedMessageCount = 0;
	private Lock lock = new Lock();
	private int sizeOfConnections;

	/**
	 * 기존에 연결된 객체를 통해서 접속한다.
	 * @param connections
	 */
	public MapReduce(ClientSide[] connections) {
		this.connections = connections;
		sizeOfConnections = connections.length;
	}

	/**
	 * 모든 대상 타겟들에 같은 메시지를 비동기로 뿌려주고, 결과를 취합한다.
	 * @param request
	 * @return
	 * @throws InterruptedException
	 */
	public Message[] map(Message request) throws InterruptedException {
		boolean[] requestable = new boolean[sizeOfConnections];
		for (int i = 0; i < requestable.length; i++)
			requestable[i] = true;
		
		return map(request, requestable);
	}

	/**
	 * 선택적인 대상 타겟들에 같은 메시지를 비동기로 뿌려주고, 결과를 취합한다.
	 * @param request
	 * @param requestable
	 * @return
	 * @throws InterruptedException
	 */
	public Message[] map(Message request, boolean[] requestable) throws InterruptedException {
		int count = 0;
		Message[] responses = new Message[sizeOfConnections];
		
		remainedMessageCount = countRequestables(requestable);
		lock.lock();
		
		for (count = 0; count < sizeOfConnections; count++) {
			if (requestable[count]) {
				MapReduceThread thread = new MapReduceThread(connections[count], request, responses, count, this);
				thread.start();
			}
			else
				responses[count] = null;
		}
		
		lock.waitForThis();
		
		return responses;
	}

	/**
	 * 대상 타겟들에, 각각의 다른 메시지들을 비동기로 뿌려주고, 결과를 취합한다.
	 * @param requests
	 * @param requestable
	 * @return
	 * @throws InterruptedException
	 */
	public Message[] map(Message[] requests, boolean[] requestable) throws InterruptedException {
		int count = 0;
		Message[] responses = new Message[sizeOfConnections];
		
		remainedMessageCount = countRequestables(requestable);
		lock.lock();
		
		for (count = 0; count < sizeOfConnections; count++) {
			if (requestable[count]) {
				MapReduceThread thread = new MapReduceThread(connections[count], requests[count], responses, count, this);
				thread.start();
			}
			else
				responses[count] = null;
		}
		lock.waitForThis();
		return responses;
	}

	/**
	 * 남은 개수를 한개 줄인다. 
	 */
	public synchronized void reduce() {
		synchronized (MapReduce.this) {
			remainedMessageCount--;
			
			//System.out.println("reduce() : " + (remainedMessageCount + 1) + " -> " + remainedMessageCount);
			
			if (remainedMessageCount == 0) {
				try {
					//System.out.println("@ unlock!!");
					lock.unlock();
				} catch (InterruptedException e) {
					e.printStackTrace();
				}
			}
		}
	}

	/**
	 * 요청을 보낼 항목의 개수를 구한다.
	 * @param requestable
	 * @return
	 */
	private int countRequestables(boolean[] requestable) {
		int count = 0;
		for (int i = 0; i < requestable.length; i++)
			if (requestable[i] == true)
				count++;
		return count;
	}
}
