package contrabase.repository;

import java.io.File;
import java.sql.Connection;
import java.sql.DriverManager;
import java.sql.PreparedStatement;
import java.sql.ResultSet;
import java.sql.SQLException;
import java.util.ArrayList;
import java.util.HashMap;
import java.util.List;
import java.util.Map;

import contrabase.ObjectContainer;
import contrabase.api.UpdateValue;
import contrabase.condition.Condition;
import contrabase.repository.datastore.ContrabaseDatastoreException;
import contrabase.repository.datastore.Tablet;
import contrabase.repository.datastore.TabletEntry;
import contrabase.repository.datastore.TabletManager;
import contrabase.repository.datastore.sqlite.SQLiteTabletManager;
import contrabase.repository.datastore.sqlite.SelectIterator;
import contrabase.util.FileUtil;

import static contrabase.condition.ConditionFactory.*;

/**
 * repository 에서 사용하는 Datastore 클래스
 */
public class Datastore {
	private TabletManager tabletManager;
	volatile private Map<String, SelectIterator> channels;
	private String dataDir;
	private volatile static int lastChannel = 0;
	
	public Datastore(String dataDir) {
		this.dataDir = dataDir;
		channels = new HashMap<String, SelectIterator>();
		this.tabletManager = new SQLiteTabletManager(dataDir);
	}
	
	// countTabletEntries
	public TabletEntries countTabletEntries() {
		TabletEntries tabletEntries = new TabletEntries();
		List<String> userId = new ArrayList<String>();
		
		// Repositry에 있는 폴더 목록(User)을 가져온다.
		File fileUserList = new File("data" + File.separator);
		String[] userList = fileUserList.list();
		
		for (String aUser : userList) {
			File user = new File("data" + File.separator + aUser);
			if (user.isDirectory())
				userId.add(aUser);
		}
		
		// 각 폴더별로 *.db(tablet) 파일 목록을 가져온다.
		// 모든 *.db 별로 key의 갯수를 구해서 TabletEntry를 생성하고 List에 add한다.
		for (String aUser : userId) {
			File fileTabletList = new File("data" + File.separator + aUser);
			String[] tabletList = fileTabletList.list();
			
			for (String aTablet : tabletList) {
				File tablet = new File("data" + File.separator + aUser + File.separator + aTablet);
				if (tablet.isFile() && aTablet.matches("^\\.db$")) {
					try {
						TabletEntry tabletEntry = getTabletEntry(aUser, aTablet);
						tabletEntries.put(tabletEntry);
					} catch (ClassNotFoundException e) {
						e.printStackTrace();
					} catch (SQLException e) {
						e.printStackTrace();
					}
				}
				
			}
		}
		
		return tabletEntries;
	}

	private TabletEntry getTabletEntry(String aUser, String aTablet) throws ClassNotFoundException, SQLException {
		String url = "jdbc:sqlite:" + dataDir + "/" + aUser + "/" + aTablet;
		FileUtil.createDirectory(dataDir);
		Class.forName("org.sqlite.JDBC");
		Connection conn = DriverManager.getConnection(url);
		String sql = "select count(key) as key_count from TABLET_KEYS";
		
		PreparedStatement pstat = conn.prepareStatement(sql);
		
		ResultSet rs = pstat.executeQuery();
		rs.next();
		int count = rs.getInt("key_count");
		TabletEntry tabletEntry = new TabletEntry(aUser, aTablet, count);
		
		return tabletEntry;
	}
	

	// createTablet
	public Tablet createTablet(String userId, String tabletName) throws ContrabaseDatastoreException {
		synchronized (this) {
			return tabletManager.createTablet(userId, tabletName);
		}
	}
	
	// deleteTablet
	public void deleteTablet(String userId, String tabletName) throws ContrabaseDatastoreException {
		synchronized (channels) {
			// ConcurrentModificationException 방지를 위해서, 복사한 후에 for 문을 수행한다.
			String[] keys = channels.keySet().toArray(new String[0]);
			
			for (String key : keys) {
				if (key.contains(merge(userId, tabletName))) {
					channels.get(key).close();
					channels.remove(key);
				}
			}
			// TODO: 여기에서, 다른 DB도 모두 닫아야 한다.
			tabletManager.deleteTablet(userId, tabletName);
		}
	}

	// existTablet
	public boolean existTablet(String userId, String tabletName) {
		synchronized (this) {
			return tabletManager.existTablet(userId, tabletName);
		}
	}

	private String merge(String userId, String tabletName) {
		return userId + "/" + tabletName;
	}

	private String merge(String userId, String tabletName, String nextId) {
		return merge(userId, tabletName) + "/" + nextId;
	}

	public String addNewChannel(String userId, String tabletName, SelectIterator values) {
		String nextId = nextChannel(userId, tabletName);
		if (values.hasNext()) {
			synchronized (channels) {
				channels.put(nextId, values);
			}
			return nextId;
		}
		else {
			return null;
		}
	}
	
	private synchronized String nextChannel(String userId, String tabletName) {
		String nextId = Long.valueOf(++lastChannel).toString();
		return merge(userId, tabletName, nextId);
	}

	public SelectIterator getChannel(String id) {
		return channels.get(id);
	}
	
	// insert
	public void insert(String userId, String tabletName, ObjectContainer value) throws ContrabaseDatastoreException {
		Tablet t = getTablet(userId, tabletName);
		t.open();
		
		try {
			t.insert(value);
		} catch (ContrabaseDatastoreException e) {
			try {
				t.close();
			} catch (ContrabaseDatastoreException ex) {
				ex.printStackTrace();
			}
			throw e;
		}
	}
	
	// update
	public int update(String userId, String tabletName, List<UpdateValue> updateValues, Condition condition) throws ContrabaseDatastoreException {
		int successed = 0;
		Tablet t = getTablet(userId, tabletName);
		t.open();
		successed = t.update(updateValues, condition);
		t.close();
		
		return successed;
	}

	// delete
	public int delete(String userId, String tabletName, Condition condition) throws ContrabaseDatastoreException {
		// delete에 성공한 갯수를 저장
		int successed = 0;
		Tablet t = getTablet(userId, tabletName);
		t.open();
		successed = t.delete(condition);
		t.close();
		
		return successed;
	}
	
	// select
	public String select(String userId, String tabletName, Condition condition) throws ContrabaseDatastoreException {
		Tablet t = getTablet(userId, tabletName);
		
		t.open();
		// select는 SelectIterator에서 conn을 사용하기 때문에, 명시적으로 닫으면 안된다!!
		SelectIterator selectIterator = t.select(condition);
		//t.close();
		
		
		String channel = null;
		
		// selection 목록에 추가해주고,
		if (selectIterator != null)
			channel = addNewChannel(userId, tabletName, selectIterator);
		
		return channel;
	}
	
	public List<ObjectContainer> selectAtOnce(String userId, String tabletName, ArrayList<String> keys) throws ContrabaseDatastoreException {
		List<ObjectContainer> list = new ArrayList<ObjectContainer>();
		Tablet t = getTablet(userId, tabletName);
		t.open();
		
		// 해당 키 목록에 대한 데이터를 모두 가져온다.
		for (String key : keys) {
			SelectIterator selectIterator = t.select(TERM("<key>", "==", key));
			
			//System.out.println(userId + "/" + tabletName + " = " + key);
			
			if (selectIterator.hasNext())
				list.add(selectIterator.next());
			
			// close() 를 해주면 안됨!! SelectIterator 내부에서 해준다!!
			//selectIterator.close();
		}
		
		t.close();
		return list;
	}
	
	// getTablet
	private Tablet getTablet(String userId, String tabletName) throws ContrabaseDatastoreException {
		return tabletManager.getTablet(userId, tabletName);
	}

	// get all keys
	public List<String> getAllKeys(String userId, String tabletName) throws ContrabaseDatastoreException {
		Tablet t = getTablet(userId, tabletName);
		t.open();
		List<String> keys = t.getAllKeys();
		t.close();
		
		return keys;
	}

	public long getRepositorySize(String userId) throws ContrabaseDatastoreException {
		long totalKeyCount = 0;

		File fileTabletList = new File("data" + File.separator + userId);
		String[] tabletList = fileTabletList.list();
		
		
		for (String aTablet : tabletList) {
			try {
				String url = "jdbc:sqlite:" + dataDir + "/" + userId + "/" + aTablet;
				FileUtil.createDirectory(dataDir);
				Class.forName("org.sqlite.JDBC");
				Connection conn = DriverManager.getConnection(url);
				String sql = "select count(key) as key_count from TABLET_KEYS";
				
				PreparedStatement pstat = conn.prepareStatement(sql);
				
				ResultSet rs = pstat.executeQuery();
				rs.next();
				int count = rs.getInt("key_count");
				totalKeyCount += count;
			} catch (ClassNotFoundException e) {
				e.printStackTrace();
			} catch (SQLException e) {
				e.printStackTrace();
			}
		}

		return totalKeyCount;
	}
}
