package contrabase.repository;


import java.io.IOException;
import java.security.NoSuchAlgorithmException;

import contrabase.ContrabaseException;
import contrabase.ContrabaseProperties;
import contrabase.PROPERTY;
import contrabase.network.ClientSide;
import contrabase.network.ContrabaseNetworkException;
import contrabase.network.MSG;
import contrabase.network.Message;
import contrabase.network.ServerSideThread;

public class RepositoryMain {
	private ClientSide messageSender = null;
	private ContrabaseProperties properties;

	private ServerSideThread repositoryThread = null;
	private ServerSideThread dataHandlerThread = null;

	private Datastore datastore;
	private String repositoryId;
	private int repositoryNo;
	private DataHandler dataHandler;
	private RepositoryGUI repositoryGUI;

	/**
	 * Repository 의 생성자
	 * 
	 * @throws ContrabaseException
	 */
	public RepositoryMain(String filename) throws ContrabaseException {
		properties = new ContrabaseProperties(filename);
		String dataDir = properties.get("data");
		datastore = new Datastore(dataDir);
		int port = Integer.valueOf(properties.get("repository.from_master.port"));
		try {
			repositoryId = RepositoryID.getRid(dataDir, port);
		} catch (NoSuchAlgorithmException e) {
			e.printStackTrace();
			throw new ContrabaseException(e);
		} catch (IOException e) {
			e.printStackTrace();
			throw new ContrabaseException(e);
		}

		// Log를 보여주는 Master GUI를 시작한다. 
		String logDir = properties.get("log");
		repositoryGUI = new RepositoryGUI(logDir);
		repositoryGUI.start();
	}

	/**
	 * Repository 클래스를 시작시킨다.
	 * 
	 * @param args 각종 옵션들..
	 */
	public static void main(String args[]) {
		try {
			RepositoryMain thisRepository = new RepositoryMain("contrabase.properties");
			thisRepository.startService();
		} catch (ContrabaseNetworkException e) {
			e.printStackTrace();
			System.exit(1);
		} catch (ContrabaseException e) {
			e.printStackTrace();
			System.exit(2);
		}
	}

	/**
	 * Repository 서비스를 시작한다.
	 * 
	 * @throws ContrabaseException
	 */
	public void startService() throws ContrabaseNetworkException {
		startDataHandler();		// API로부터 select 를 처리할 핸들러를 실행한다.
		startMainHandler();		// Master 의 요청을 받는 메인 핸들러를 실행하고,
		connectToMaster();		// Master 에 요청을 보낼 채널을 만든다.
	}

	/**
	 * Client API 로부터 select 요청을 처리할 핸들러를 시작한다.
	 * 
	 * @throws ContrabaseNetworkException
	 */
	private void startDataHandler() throws ContrabaseNetworkException {
		int port = Integer.valueOf(properties.get(PROPERTY.REPOSITORY_FROM_CLIENT_PORT));
		dataHandler = new DataHandler(datastore);
		dataHandlerThread = new ServerSideThread(port, dataHandler);
		dataHandlerThread.start();
	}

	/**
	 * Repository 의 메인 핸들러를 시작한다.
	 * 
	 * @throws ContrabaseNetworkException
	 */
	private void startMainHandler() throws ContrabaseNetworkException {
		int port = Integer.valueOf(properties.get(PROPERTY.REPOSITORY_FROM_MASTER_PORT));
		// repositoryThread = new ServerSideThread(port, new RequestHandler(datastore, repositoryId, properties));
		repositoryThread = new ServerSideThread(port, new RequestHandler(datastore, repositoryId, properties, repositoryGUI));
		repositoryThread.start();
	}

	/**
	 * Master 서버에 접속한다.
	 * 
	 * @throws ContrabaseNetworkException
	 */
	private void connectToMaster() throws ContrabaseNetworkException {
		// 설정 파일에서, 서버의 IP 와 port 를 들고 와서..
		String masterIP = properties.get(PROPERTY.REPOSITORY_TO_MASTER_IP);
		int masterPort = Integer.valueOf(properties.get(PROPERTY.REPOSITORY_TO_MASTER_PORT));
		int handlerPort = Integer.valueOf(properties.get(PROPERTY.REPOSITORY_FROM_MASTER_PORT));
		int dataHandlerPort = Integer.valueOf(properties.get(PROPERTY.REPOSITORY_FROM_CLIENT_PORT));
		
		// 서버에 접속한다.
		messageSender = new ClientSide(masterIP, masterPort);
		
		Message msg = new Message();
		
		// 연결 초기화
		msg.set(MSG.ACTION, MSG.MANAGEMENT.CONNECTION_INIT);
		
		String localIP = messageSender.getLocalIP();
		
		// 서버에, 리포지토리의 RID와 핸들러의 IP와 포트를 보내준다.
		msg.set(MSG.MANAGEMENT.REPOSITORY_RID, repositoryId);
		msg.set(MSG.MANAGEMENT.REPOSITORY_IP, localIP);
		msg.set(MSG.MANAGEMENT.REPOSITORY_PORT, handlerPort);
		
		// 데이터 핸들러의 IP와 port 번호.
		msg.set(MSG.MANAGEMENT.REPOSITORY_DATA_IP, localIP);
		msg.set(MSG.MANAGEMENT.REPOSITORY_DATA_PORT, dataHandlerPort);
		
		// Tablet 들의 정보를 보낸다.
		msg.set(MSG.MANAGEMENT.TABLET_ENTRIES, datastore.countTabletEntries());
		
		// 전송한 후에,
		Message response = messageSender.send(msg);
		
		// 접속 요청이 에러면, 뭔가 잘못 되었음.
		if (response.isFailed())
			throw new RuntimeException("fatal error : never reached statement here.");
		
		// 정상적으로 접속 되었으면, repositoryNo 를 설정 한다.
		repositoryNo = (Integer)response.get(MSG.MANAGEMENT.REPOSITORY_NO);
		dataHandler.setRepositoryNo(repositoryNo);
		dataHandler.setRepositoryId(repositoryId);
		
	}
}
