package contrabase.repository;

import java.net.InetAddress;
import java.net.UnknownHostException;
import java.util.ArrayList;
import java.util.HashMap;
import java.util.List;
import java.util.Map;

import contrabase.ContrabaseProperties;
import contrabase.ObjectContainer;
import contrabase.PROPERTY;
import contrabase.api.UpdateValue;
import contrabase.condition.Condition;
import contrabase.network.ClientSide;
import contrabase.network.ContrabaseNetworkException;
import contrabase.network.MSG;
import contrabase.network.Message;
import contrabase.network.MessageHandler;
import contrabase.repository.datastore.ContrabaseDatastoreException;
import contrabase.util.Pair;
import static contrabase.condition.ConditionFactory.*;

class RequestHandler implements MessageHandler {
	private Datastore datastore;
	private String rid;
	private ContrabaseProperties properties;
	private RepositoryGUI repositoryGUI;

	public RequestHandler(Datastore datastore, String rid, ContrabaseProperties properties) {
		this.datastore = datastore;
		this.rid = rid;
		this.properties = properties;
	}

	public RequestHandler(Datastore datastore, String rid, ContrabaseProperties properties, RepositoryGUI repositoryGUI) {
		this.datastore = datastore;
		this.rid = rid;
		this.properties = properties;
		this.repositoryGUI = repositoryGUI;
	}

	@Override
	public void handle(Message request, Message response) {
		String action = (String)request.get(MSG.ACTION);
		String userId = (String)request.get(MSG.REQUEST.USER_ID);
		String tabletName = (String)request.get(MSG.REQUEST.TABLE_NAME);
		
		if (action == null) {
			response.setFailed("Null Action");
			action = "Null Action";
		}
		
		else if (action.equals(MSG.REQUEST.CREATE_TABLE))
			handleCreateTable(request, response, userId, tabletName);
		
		else if (action.equals(MSG.REQUEST.DELETE_TABLE))
			handleDeleteTable(request, response, userId, tabletName);
		
		else if (action.equals(MSG.REQUEST.INSERT))
			handleInsert(request, response, userId, tabletName);
		
		else if (action.equals(MSG.REQUEST.UPDATE))
			handleUpdate(request, response, userId, tabletName);
		
		else if (action.equals(MSG.REQUEST.DELETE))
			handleDelete(request, response, userId, tabletName);
		
		else if (action.equals(MSG.REQUEST.SELECT))
			handleSelect(request, response, userId, tabletName);
		
		else if (action.equals(MSG.MANAGEMENT.GET_KEYS))
			handleGetKeys(request, response, userId, tabletName);
		
		else if (action.equals(MSG.REQUEST.LOAD_BALANCE_TABLE))
			handleLoadBalancing(request, response);
		
		else if (action.equals(MSG.REQUEST.GET_REPOSITORY_SIZE))
			handleGetRepositorySize(request, response, userId);
		
		else if (action.equals(MSG.REQUEST.GET_REPOSITORY_MEMORY))
			handleGetRepositoryMemory(request, response);

		else {
			handleUndefined(request, response);
		}
		
		// 로그 기록
		repositoryGUI.appendLog(userId + " : " + action);
		
		// 마스터에 보내는 모든 응답에, RID 를 써준다.
		response.set(MSG.RESPONSE.RID, rid);
	}

	/**
	 * 동기화 요청을 처리한다.
	 * @param request
	 * @param response
	 */
	@SuppressWarnings("unchecked")
	private void handleLoadBalancing(Message request, Message response) {
		String userId = (String)request.get(MSG.REQUEST.USER_ID);
		String tabletName = (String)request.get(MSG.REQUEST.TABLE_NAME);
		
		// ArrayList<Pair< KEY, 받을놈RID >>
		ArrayList<Pair<String, String>> diff = (ArrayList<Pair<String, String>>)request.get(MSG.MANAGEMENT.LOAD_BALANCE_DATA);
		
		// Map< RID, Pair< IP, Port > >
		Map<String, Pair<String, Integer>> addresses = (Map<String, Pair<String, Integer>>)request.get(MSG.MANAGEMENT.REPOSITORY_ADDRESSES);
		
		// 로드 밸런싱 한다!!!!
		Map<String, ArrayList<String>> data = collectData(diff);
		for (Map.Entry<String, ArrayList<String>> entry : data.entrySet()) {
			String targetRid = entry.getKey();
			ArrayList<String> list = entry.getValue();
			
			// targetRid 에다가, list 에 해당하는 데이터를 날린다!!!
			Pair<String, Integer> addr = addresses.get(targetRid);
			try {
				sendData(addr.getLeft(), addr.getRight(), userId, tabletName, list);
			} catch (ContrabaseNetworkException e) {
				e.printStackTrace();
				response.setFailed(e.getLocalizedMessage());
				return;
			} catch (ContrabaseDatastoreException e) {
				e.printStackTrace();
				response.setFailed(e.getLocalizedMessage());
				return;
			}
		}
		
		// 성공.
		response.setSuccessed();
	}

	private void sendData(String ip, int port, String userId, String tabletName, ArrayList<String> list) throws ContrabaseNetworkException, ContrabaseDatastoreException {
		ClientSide client = new ClientSide(ip, port);
		Message msg = new Message();
		
		// 키 목록을 가지고, 모든 데이터를 <한번에> 가져온다.
		List<ObjectContainer> data = datastore.selectAtOnce(userId, tabletName, list);
		
		msg.set(MSG.ACTION, MSG.MANAGEMENT.LOAD_BALANCE_ACTION);
		msg.set(MSG.REQUEST.USER_ID, userId);
		msg.set(MSG.REQUEST.TABLE_NAME, tabletName);
		msg.set(MSG.MANAGEMENT.LOAD_BALANCE_DATA, data);
		
		Message response = client.send(msg);
		
		if (response.isFailed())
			throw new RuntimeException(response.getErrorMessage());
		else {
			// 성공했으면, 해당 목록에 있는 키의 데이터들을 모두 지운다.
			for (String key : list)
				datastore.delete(userId, tabletName, TERM("<key>", "==", key));
		}
		
		client.close();
	}

	/**
	 * 목록의 구성을 변환한다.
	 * ArrayList<Pair<String, String>> ===>> Map<String, ArrayList<String>>
	 * ArrayList<Pair< KEY  , 타겟RID>> ===>> Map<타겟RID, ArrayList< KEY  >>  
	 * @param diff
	 * @return
	 */
	private Map<String, ArrayList<String>> collectData(ArrayList<Pair<String, String>> diff) {
		Map<String, ArrayList<String>> map = new HashMap<String, ArrayList<String>>();
		
		for (Pair<String, String> entry : diff) {
			String key = entry.getLeft();
			String targetRid = entry.getRight();
			
			ArrayList<String> list = null;
			if (map.containsKey(targetRid))
				list = map.get(targetRid);
			else {
				list = new ArrayList<String>();
				map.put(targetRid, list);
			}
			
			list.add(key);
		}
		
		return map;
	}

	private void handleGetKeys(Message request, Message response, String userId, String tabletName) {
		try {
			List<String> keys = datastore.getAllKeys(userId, tabletName);
			
			response.set(MSG.RESPONSE.KEY_LIST, keys);
			response.setSuccessed();
		} catch (ContrabaseDatastoreException e) {
			response.setFailed("\"getAllKeys\" Failed");
			e.printStackTrace();
		}
	}

	private void handleUndefined(Message request, Message response) {
		String action = (String)request.get(MSG.ACTION);
		response.setFailed("Undefined action: " + action);
	}

	private void handleSelect(Message request, Message response, String userId, String tabletName) {
		try {
			// Channel 을 받아서.
			String channel = datastore.select(userId, tabletName, (Condition)request.get(MSG.REQUEST.CONDITION));
			
			response.setSuccessed();
			
			if (channel != null) {
				response.set(MSG.RESPONSE.HAS_SELECTION_DATA, true);
				response.set(MSG.RESPONSE.REPOSITORY_CHANNEL, channel);
				
				// repository IP, Port setting
				String ip = InetAddress.getLocalHost().getHostAddress();
				String port = properties.get(PROPERTY.REPOSITORY_FROM_CLIENT_PORT);
				response.set(MSG.RESPONSE.REPOSITORY_IP, ip);
				response.set(MSG.RESPONSE.REPOSITORY_PORT, Integer.valueOf(port));
				
				// setting rid
				response.set(MSG.RESPONSE.RID, rid);
			}
			else {
				response.set(MSG.RESPONSE.HAS_SELECTION_DATA, false);
			}
		} catch (UnknownHostException e) {
			response.setFailed("\"select\" Failed");
			e.printStackTrace();
		} catch (ContrabaseDatastoreException e) {
			response.setFailed("\"selete\" Failed");
			e.printStackTrace();
		}
	}

	private void handleDelete(Message request, Message response, String userId, String tabletName) {
		// delete 성공 갯수.
		try {
			int count = datastore.delete(userId, tabletName, (Condition)request.get(MSG.REQUEST.CONDITION));
			
			response.set(MSG.RESPONSE.COUNT, count);
			response.setSuccessed();
		} catch (ContrabaseDatastoreException e) {
			response.setFailed("\"delete\" Failed");
			e.printStackTrace();
		}
	}

	@SuppressWarnings("unchecked")
	private void handleUpdate(Message request, Message response, String userId, String tabletName) {
		try {
			// update 성공 갯수.
			int count = datastore.update(
				userId,
				tabletName,
				(List<UpdateValue>)request.get(MSG.REQUEST.UPDATE_VALUES),
				(Condition)request.get(MSG.REQUEST.CONDITION)
			);
			response.set(MSG.RESPONSE.COUNT, count);
			// success
			response.setSuccessed();
		} catch (ContrabaseDatastoreException e) {
			response.setFailed("\"update\" Failed");
			e.printStackTrace();
		}
	}

	private void handleInsert(Message request, Message response, String userId, String tabletName) {
		try {
			datastore.insert(userId, tabletName, (ObjectContainer)request.get(MSG.REQUEST.INSERT_DATA));
			// success
			response.setSuccessed();				
		} catch (ContrabaseDatastoreException e) {
			response.setFailed("\"insert\" Failed");
			e.printStackTrace();
		}
	}

	private void handleDeleteTable(Message request, Message response, String userId, String tabletName) {
		try {
			datastore.deleteTablet(userId, tabletName);
			// success
			response.setSuccessed();
		} catch (ContrabaseDatastoreException e) {
			response.setFailed("\"delete Table\" Failed");
			e.printStackTrace();
		}
	}

	private void handleCreateTable(Message requset, Message response, String userId, String tabletName) {
		try {
			datastore.createTablet(userId, tabletName);
			response.setSuccessed();
		} catch (ContrabaseDatastoreException e) {
			response.setFailed("\"create Table\" Failed");
			e.printStackTrace();
		}
	}

	private void handleGetRepositorySize(Message request, Message response, String userId) {
		// Repository에 존재하는 모든 Tablet의 Key Count를 종합한다.

		try {
			long size = datastore.getRepositorySize(userId);
			
			response.set(MSG.RESPONSE.ACTION_DATA, size);
			response.setSuccessed();
		} catch (ContrabaseDatastoreException e) {
			response.setFailed("\"getRepositorySize\" Failed");
			e.printStackTrace();
		}
	}

	private void handleGetRepositoryMemory(Message request, Message response) {
		// Memory 사용량
		long memory = Runtime.getRuntime().totalMemory()- Runtime.getRuntime().freeMemory();
		
		response.set(MSG.RESPONSE.ACTION_DATA, memory);
		response.setSuccessed();
	}
}
