package contrabase.repository.datastore.sqlite;

import java.sql.DriverManager;
import java.sql.PreparedStatement;
import java.sql.ResultSet;
import java.sql.SQLException;
import java.sql.Connection;
import java.sql.Statement;
import java.util.ArrayList;
import java.util.Iterator;
import java.util.List;
import java.util.SortedSet;

import contrabase.ObjectContainer;
import contrabase.ObjectContainerField;
import contrabase.api.UpdateValue;
import contrabase.condition.Condition;
import contrabase.repository.datastore.ContrabaseDatastoreException;
import contrabase.repository.datastore.Tablet;
import contrabase.repository.datastore.sqlite.update.SQLiteUpdateValue;


/**
 * @author Shiva83
 *
 * SQLite를 이용해서 Tablet를 관리하기 위한 객체
 */
public class SQLiteTablet implements Tablet {
	private final static String[] tableNames = { "TABLET_NUMBER", "TABLET_TEXT", "TABLET_BLOB", "TABLET_BOOLEAN" };
	private Connection conn;
	private String dataDir;
	private String userId;
	private String tabletName;
	
	/**
	 * @param userId Tablet을 사용하기 위한 ID
	 * @param tabletName Tablet이 각 Repository에
	 *        분할 되어진 Tablet Name. 기본적으로 Table Name과 같다.
	 * @throws ContrabaseDatastoreException
	 */
	public SQLiteTablet(String dataDir, String userId, String tabletName) throws ContrabaseDatastoreException {
		this.dataDir = dataDir;
		this.userId = userId;
		this.tabletName = tabletName;
	}

	@Override
	public void open() throws ContrabaseDatastoreException {
		// Connection 객체가 Null일 때만 Class의 Member변수를 이용하여
		// Connection을 생성한다.
		if (conn == null) {
			// url : data/userId/tableName.db
			String url = "jdbc:sqlite:" + dataDir + "/" + userId + "/" + tabletName + ".db";
			try {
				Class.forName("org.sqlite.JDBC");
				conn = DriverManager.getConnection(url);
			} catch (ClassNotFoundException e) {
				e.printStackTrace();
			} catch (SQLException e) {
				e.printStackTrace();
			}
		} // end if
	}
	
	@Override
	public void close() throws ContrabaseDatastoreException {
		try {
			if (conn != null)
				conn.close();
		} catch (SQLException e) {
			throw new ContrabaseDatastoreException("SQLException Close Error", e);
		}
	}
	
	public void finalize() {
		try {
			// 만약에 명시적으로 close를 하지 않았을 경우를 대비.
			close();
		} catch (ContrabaseDatastoreException e) {
			e.printStackTrace();
		} finally {
			try {
				super.finalize();
			} catch (Throwable e) {
				e.printStackTrace();
			}
		}
	}
	
	@Override
	public int delete(Condition condition) throws ContrabaseDatastoreException {
		// condition에 맞는 객체를 삭제한 갯수를 저장
		int successed = 0;
		try {
			Statement stat = conn.createStatement();
			
			// 입력받은 조건(condition)을 이용하여 delete할 객체의 keys를 갖고 온다.
			SortedSet<String> keys = condition.getKeys(new SQLiteKeySelector(stat));
			
			// keys를 이용하여 Tablet에 저장된 객체를 삭제한다.
			// condition에 맞는 key가 있을때만 successed가 증가한다.
			for (String key : keys) {
				try {
					stat.execute("delete from TABLET_KEYS where key = '" + key + "'");
					stat.execute("delete from TABLET_NUMBER where key = '" + key + "'");
					stat.execute("delete from TABLET_TEXT where key = '" + key + "'");
					stat.execute("delete from TABLET_BLOB where key = '" + key + "'");
					stat.execute("delete from TABLET_BOOLEAN where key = '" + key + "'");
					successed++;
				}
				catch (SQLException ex) {
					// do nothing : 해당 key 를 삭제 하지 못해도, 다음 항목을 삭제 하기 위해서 try-catch 를 해준다.
					ex.printStackTrace();
				}
			}
			stat.close();
		} catch (SQLException e) {
			throw new ContrabaseDatastoreException("Exception Delete Error", e);
		}
		
		// 총 삭제한 갯수를 return한다.
		// conditon에 맞는 key가 하나도 없으면 successed의 갯수는 0개
		
		return successed;
	}

	@Override
	public void insert(ObjectContainer value) throws ContrabaseDatastoreException {
		try {
			// ObjectContainer에 저장되어 있는 key를 가져오고.
			String key = value.getKey();
			// ObjectContainer에 저장되어 있는 각 field(ObjectContainerField)를 Iterator에 담아서.
			Iterator<ObjectContainerField> fields = value.iterator();
			Statement stat = conn.createStatement();
			try {
				// ObjectContainer에 저장되어 있는 Class와 Key를 저장하고.
				stat.execute("insert into TABLET_KEYS values('" + key + "', '" + value.getClassName() + "')");
			} catch (Exception e) {
				throw new ContrabaseDatastoreException("Exist Key value = " + key, e);
			}
			
			while (fields.hasNext()) {
				ObjectContainerField field = fields.next();
				Object objv = field.getValue();
				String sql = "insert into TABLET_";

				// 각각의 알맞는 Tablet에 field를 저장한다.
				if (objv instanceof Number) {
					sql += "NUMBER values('" + key + "', '" + field.getTypeName() 
												+ "', '" + field.getFieldName()
												+ "', " + objv.toString();
				}
				else if (objv instanceof Character
						|| objv instanceof String) {
					sql += "TEXT values('" + key + "', '" + field.getTypeName()
												 + "', '" + field.getFieldName()
												 + "', '" + objv.toString() + "'";
				}
				else if (objv instanceof Boolean) {
					// boolean의 경우 true/false를 직접 저장하지 않고 
					// true일 경우 1로 false일 경우 0으로 저장한다.
					int bool = SQLiteUtil.boolStringToInt(objv.toString());
					sql += "BOOLEAN values('" + key + "', '" + field.getTypeName()
													+ "', '" + field.getFieldName()
													+ "', " + bool;
				}
				else {
					// Blob의 경우 value를 byte로 저장한다.
					PreparedStatement pstat = conn.prepareStatement("insert into tablet_BLOB values(?, ?, ?, ?)");
					pstat.setString(1, key);
					pstat.setString(2, field.getTypeName());
					pstat.setString(3, field.getFieldName());
					pstat.setBytes(4, (byte[])field.getValue());
					pstat.execute();
					continue;
				}
				sql += ")";
				stat.execute(sql);
				stat.close();
			} // end of while
		
		} catch (SQLException e) {
			throw new ContrabaseDatastoreException("SQL Exception Insert Error", e);
		}
		
		//return true;
	}

	@Override
	public SelectIterator select(Condition condition) throws ContrabaseDatastoreException {
		try {
			Statement stat = conn.createStatement();
			
			// 입력받은 조건(condition)을 이용하여 select할 객체의 keys를 갖고 온다.
			SortedSet<String> keys = condition.getKeys(new SQLiteKeySelector(stat));
			
			// Connection객체와 keys의 iterator를. SelectIterator의 생성자로 넘겨주고,
			// 그 SelectIterator를 이용하여 Select를 한다.
			return new SelectIterator(keys.iterator(), conn);
			
		} catch (SQLException e) {
			e.printStackTrace();
		}
		
		return null;
	}

	@Override
	public int update(List<UpdateValue> updateValues, Condition condition) throws ContrabaseDatastoreException {
		int successed = 0;
		
		try {
			Statement stat = conn.createStatement();
			
			// 입력받은 조건(condition)을 이용하여 update할 객체의 keys를 갖고 온다.
			SortedSet<String> keys = condition.getKeys(new SQLiteKeySelector(stat));
			
			stat.close();
			
			for (String key : keys) {
				for (UpdateValue field : updateValues) {
					// DB를 뒤져서, 이 key 의 field 의 타입이름을 String 으로 통째로 들고와서 typeName 에 대입한다.
					String typeName = getTypename(key, field);
					
					// 만약, 이 key 에 해당 fieldName 이 존재하지 않으면, 그냥, continue; 한다. (안쪽의 for문을 제끼고 다음 key로 넘어감.)
					if (typeName == null)
						continue;
					
					// 매우 특수한 경우인데, DB에 있는 컬럼의 타입이 char 인데, 2글자 이상을 대입하면 에러.
					if (typeName.equals("java.lang.Character")
						&& field.isText()
						&& field.getValueString().length() != 1) {
						throw new ContrabaseDatastoreException("not a Character: " + field.getValueString());
					}
					
					// 해당하는 타입이름에 해당하는, 적절한 객체를 생성한다.
					SQLiteUpdateValue fieldUpdater = SQLiteUpdateValue.createSQLiteUpdateValue(conn, typeName);
					// 적절한 객체로, 해당 필드를 업데이트 하고, 성공여부를 확인한다.
					if (fieldUpdater.updateValue(key, field))
						successed++;
					
				} // end of for (UpdateValue field : updateValues)
			} // end of for (String key : keys)
		} catch (SQLException e) {
			throw new ContrabaseDatastoreException("SQL Exception Update Error", e);
		}
		
		return successed;
	}
	
	@Override
	public List<String> getAllKeys() throws ContrabaseDatastoreException {
		List<String> keys = new ArrayList<String>();
		Statement stat = null;
		
		try {
			stat = conn.createStatement();
			ResultSet rs = stat.executeQuery("select key from TABLET_KEYS");
			
			while (rs.next())
				keys.add(rs.getString("key"));
			
			rs.close();
			stat.close();
		} catch (SQLException e) {
			throw new ContrabaseDatastoreException(e);
		}
		
		return keys;
	}
	
	private String getTypename(String key, UpdateValue field) throws SQLException {
		Statement stat = conn.createStatement();
		ResultSet rs = null;
		String typeName = null;

		for (String table : tableNames) {
			// TABLET_???(key TEXT, type_name TEXT, var_name TEXT, value ????)
			rs = stat.executeQuery("select type_name from " + table + " where key = '" + key + "' and var_name = " + field.getFieldName());
			if (rs.next()) {
				typeName = rs.getString(1);
				break;
			}
		}

		stat.close();
		return typeName;
	}
}
