package contrabase.repository.datastore.sqlite;

import java.io.File;
import java.sql.Connection;
import java.sql.DriverManager;
import java.sql.PreparedStatement;
import java.sql.ResultSet;
import java.sql.SQLException;
import java.sql.Statement;

import contrabase.repository.datastore.ContrabaseDatastoreException;
import contrabase.repository.datastore.Tablet;
import contrabase.repository.datastore.TabletManager;
import contrabase.util.FileUtil;

/**
 * Tablet 의 생성(create), 삭제(delete), 연결(get)을 관리하는 객체
 * @author Shiva83
 *
 */
public class SQLiteTabletManager implements TabletManager {
	private Tablet tablet;
	private String dataDir;
	
	public SQLiteTabletManager(String dataDir) {
		this.dataDir = dataDir;
	}
	
	public Tablet createTablet(String userId, String tabletName) throws ContrabaseDatastoreException {
		FileUtil.createDirectory(dataDir + File.separator + userId);
		
		// url : data/userId/tableName.db
		String url = "jdbc:sqlite:" + dataDir + "/" + userId + "/" + tabletName + ".db";
		try {
			Class.forName("org.sqlite.JDBC");
			Connection conn = DriverManager.getConnection(url);
			Statement stat = conn.createStatement();
			
			// POJO를 저장하기 위해 5개의 Tablet Table을 생성한다.
			// TABLET_KEYS는. key와 저장되는 class의 name을 저장한다.
			stat.execute("create table TABLET_KEYS(key TEXT, class_name TEXT, primary key(key, class_name))");
			// TABLET_NUMBER는 key와 변수의 Type Name(int, byte, short, long, float, double
			// 										  Integer, Byte, Short, Long, Float, Double), Variable Name, value를 저장한다.
			stat.execute("create table TABLET_NUMBER(key TEXT, type_name TEXT, var_name TEXT, value REAL)");
			// TABLET_TEXT는 key와 변수의 Type Name(char, Character, String), Variable Name, value를 저장한다.
			stat.execute("create table TABLET_TEXT(key TEXT, type_name TEXT, var_name TEXT, value TEXT)");
			// TALBET_BLOB는 key와 변수의 Type Name
			stat.execute("create table TABLET_BLOB(key TEXT, type_name TEXT, var_name TEXT, value BLOB)");
			stat.execute("create table TABLET_BOOLEAN(key TEXT, type_name TEXT, var_name TEXT, value INTEGER)");
			
			stat.close();
			conn.close();
		} catch (ClassNotFoundException e) {
			throw new ContrabaseDatastoreException("org.sqlite.JDBC not found", e);
		} catch (SQLException e) {
			throw new ContrabaseDatastoreException("SQLException occur", e);
		}
		
		return getTablet(userId, tabletName);
	}
	
	public Tablet getTablet(String userId, String tabletName) throws ContrabaseDatastoreException {
		tablet = new SQLiteTablet(dataDir, userId, tabletName);
		return tablet;
	}
	
	public void deleteTablet(String userId, String tabletName) throws ContrabaseDatastoreException {
		File f = new File(dataDir + "/" + userId + "/" + tabletName + ".db");
		f.delete();
	}
	
	public boolean existTablet(String userId, String tabletName) {
		String url = "jdbc:sqlite:" + dataDir + "/meta.db";
		
		try {
			Class.forName("org.sqlite.JDBC");
			Connection conn = DriverManager.getConnection(url);
			PreparedStatement pstat = conn.prepareStatement("select Id, TabletName from MetaTablet where Id = ? and TabletName = ?");
			pstat.setString(1, userId);
			pstat.setString(2, tabletName);
			
			ResultSet rs = pstat.executeQuery();
			rs.next();
			
			if(userId.equals(rs.getString(1)) && tabletName.equals(rs.getString(2)))
				return true;
		} catch (ClassNotFoundException e) {
			e.printStackTrace();
		} catch (SQLException e) {
			e.printStackTrace();
		}
		
		return false;
	}
}