package contrabase.repository.datastore.sqlite;

import java.sql.Connection;
import java.sql.PreparedStatement;
import java.sql.ResultSet;
import java.sql.SQLException;
import java.util.Iterator;
import java.util.NoSuchElementException;

import contrabase.ObjectContainerField;
import contrabase.ObjectContainer;

import contrabase.util.Converter;

/**
 * @author Shiva83
 *
 * Select를 할 때 실제로는 
 * SelectItarator 객체의 next() Method를 이용하여 
 * 한 번에 하나의 ObjectContainer를 return 받아 사용한다.
 */
public class SelectIterator {
	Iterator<String> keys;
	Connection conn;
	
	/**
	 * @param keys 생성자로 Iterator 객체로 key의 목록을 갖고 있다.
	 * @param conn 생성자로 넘겨받은 Connection. Select한 Tablet과의 연결을 유지하고 있다. 
	 */
	public SelectIterator(Iterator<String> keys, Connection conn) {
		this.keys = keys;
		this.conn = conn;
	}

	public ObjectContainer next() {
		String key = null;
		
		// keys Iterator를 이용하여 Select할 객체가 남아 있는지 확인한다.
		// 있다면 key에 저장하고
		if (this.keys.hasNext())
			key = keys.next();
		
		// 없다면 RuntimeException
		else {
			close();
			throw new NoSuchElementException("Not Exist more key");
		}
		
		
		ObjectContainer oc = new ObjectContainer();
		
		// 클래스 이름을 채우고,
		fillClassName(key, oc);
		
		// 객체 내용을 채운다.
		try {
			oc = fillObjectContainer(key, oc);
		} catch (SQLException e) {
			e.printStackTrace();
		}
		
		return oc;
	}

	private void fillClassName(String key, ObjectContainer oc) {
		ResultSet rs = null;
		PreparedStatement pstat = null;
		try {
			// TABLET_KEYS 에서 해당 key의 Class Name을 가져온뒤
			String sql = "SELECT class_name FROM TABLET_KEYS WHERE key = ?";
			
			pstat = conn.prepareStatement(sql);
			pstat.setString(1, key);
			rs = pstat.executeQuery();
			
			if (rs.next()) {
				String className = rs.getString("class_name");
				oc.setClassName(className);
			}
		} catch (SQLException e) {
			e.printStackTrace();
		} finally {
			try {
				if (rs != null)
					rs.close();
				if (pstat != null)
					pstat.close();
			} catch (SQLException e) {
				e.printStackTrace();
			}
		}
	}

	private ObjectContainer fillObjectContainer(String key, ObjectContainer oc) throws SQLException {
		// ObjectContainer에 Key를 입력해준다.
		oc.setKey(key);
		
		{
			// NUMBER 에서, key에 해당하는 row를 가져와서 루프를 돌면서, fields[] 안에 있는거면, 추가.
			// TABLET_NUMBER(key TEXT, type_name TEXT, var_name TEXT, value REAL).
			String sql = "select * from TABLET_NUMBER where key = ?";
			PreparedStatement pstat = conn.prepareStatement(sql);
			pstat.setString(1, key);
			
			ResultSet rs = pstat.executeQuery();
			while (rs.next()) {
				String typeName = rs.getString(2);
				oc.add( new ObjectContainerField(rs.getString(3), typeName, Converter.doubleToObject(typeName, rs.getDouble(4))) );
			}
			
			pstat.close();
		}
		
		{
			// TEXT   에서, key에 해당하는 row를 가져와서 루프를 돌면서, fields[] 안에 있는거면, 추가.
			// TABLET_TEXT(key TEXT, type_name TEXT, var_name TEXT, value TEXT).
			String sql = "select * from TABLET_TEXT where key = ?";
			PreparedStatement pstat = conn.prepareStatement(sql);
			pstat.setString(1, key);
			
			ResultSet rs = pstat.executeQuery();
			while (rs.next()) {
				oc.add( new ObjectContainerField(rs.getString(3), rs.getString(2), rs.getString(4)) );
			}
			
			pstat.close();
		}
		
		{
			// BOOLEAN에서, key에 해당하는 row를 가져와서 루프를 돌면서, fields[] 안에 있는거면, 추가.
			// TABLET_BOOLEAN(key TEXT, type_name TEXT, var_name TEXT, value INTEGER)
			String sql = "select * from TABLET_BOOLEAN where key = ?";
			PreparedStatement pstat = conn.prepareStatement(sql);
			pstat.setString(1, key);
			
			ResultSet rs = pstat.executeQuery();
			while (rs.next()) {
				boolean bool = true;
				if (rs.getInt(4) == 0)
					bool = false;
				oc.add( new ObjectContainerField(rs.getString(3), rs.getString(2), bool) );
			}
			
			pstat.close();
		}
		
		{
			// BLOB   에서, key에 해당하는 row를 가져와서 루프를 돌면서, fields[] 안에 있는거면, 추가.
			// TABLET_BLOB(key TEXT, type_name TEXT, var_name TEXT, value BLOB)
			String sql = "select * from TABLET_BLOB where key = ?";
			PreparedStatement pstat = conn.prepareStatement(sql);
			pstat.setString(1, key);
			
			ResultSet rs = pstat.executeQuery();
			while (rs.next()) {
				oc.add( new ObjectContainerField(rs.getString(3), rs.getString(2), rs.getBytes(4)) );
			}
			
			pstat.close();
		}
		
		return oc;
	}
	
	public void close() {
		if (conn != null) {
			try {
				conn.close();
			} catch (SQLException e) {
				e.printStackTrace();
			}
		}
	}
	
	public boolean hasNext() {
		// keys에 다음 key가 남아 있는지 확인.
		boolean has = keys.hasNext();
		if (has == false)
			close();
		return has;
	}
}
