package contrabase.repository.datastore.sqlite.update;

import java.sql.Connection;
import java.sql.PreparedStatement;
import java.sql.SQLException;
import java.sql.Statement;

import contrabase.api.UpdateValue;

/**
 * SQLite 에서 UpdateValue 를 사용하기 위한 추상클래스
 * @author jong10
 */
public abstract class SQLiteUpdateValue {
	private Connection connection;

	/**
	 * 추상클래스
	 * @param key 키
	 * @param value 업데이트 정보를 담고 있는 객체
	 * @return 성공인지 실패인지
	 */
	public abstract boolean updateValue(String key, UpdateValue value);

	/**
	 * 상속한 클래스에서만 사용한다.
	 * @param connection 이미 연결된 SQL 커넥션
	 */
	protected void setConnection(Connection connection) {
		this.connection = connection;
	}

	/**
	 * 타입 이름에 해당하는 객체를 생성한다. Factory Method Pattern 이용.
	 * @param conn
	 * @param typeName
	 * @return 해당하는 객체가 리턴된다.
	 */
	public static SQLiteUpdateValue createSQLiteUpdateValue(Connection conn, String typeName) {
		if (typeName == null || typeName.length() == 0)
			throw new RuntimeException("typeName is null or emtpy");
		
		// TEXT
		if (typeName.equals("java.lang.Character") ||
			typeName.equals("char") ||
			typeName.equals("java.lang.String")) {
			return new SQLiteUpdateValueText(conn);
		}
		
		// DECIMAL
		else if (typeName.equals("java.lang.Byte") ||
				typeName.equals("byte") ||
				typeName.equals("java.lang.Short") ||
				typeName.equals("short") ||
				typeName.equals("java.lang.Integer") ||
				typeName.equals("int") ||
				typeName.equals("java.lang.Long") ||
				typeName.equals("long")) {
			return new SQLiteUpdateValueDecimal(conn);
		}
		
		// REAL
		else if (typeName.equals("java.lang.Float") ||
				typeName.equals("float") ||
				typeName.equals("java.lang.Double") ||
				typeName.equals("double")) {
			return new SQLiteUpdateValueReal(conn);
		}
		
		// BOOLEAN
		else if (typeName.equals("java.lang.Boolean") ||
				typeName.equals("boolean")) {
			return new SQLiteUpdateValueBoolean(conn);
		}
		
		// 그 밖의 객체 (BLOB)
		else {
			return new SQLiteUpdateValueBlob(conn);
		}
	}

	/**
	 * 실제 해당 테이블에서, 해당 필드의 값을, (문자열로 바꾼) 해당 값으로 업데이트하는 메소드.
	 * 이 메소드는 구상클래스에서 사용한다.
	 * @param tableName SQLite 에서 사용하는 테이블 이름
	 * @param formattedField 대입하려는 필드의 이름을 문자열로 바꾼 형태
	 * @param formattedValue 값을 문자열로 바꾼 형태. (ex: 'name', 1234, 3.14, ...)
	 * @return 성공여부
	 */
	protected boolean updateField(String tableName, String formattedField, String formattedValue, String key) {
		Statement stat = null;
		int successed = 0;
		
		try {
			stat = connection.createStatement();
			
			key = "'" + key + "'";
			
			// UPDATE {tableName}
			// SET {field} = {formattedValue}
			// WHERE var_name = {field} and 
			String sql = "UPDATE " + tableName
				+ " SET VALUE = " + formattedValue
				+ " WHERE var_name = " + formattedField + " AND key = " + key;
			
			// 결과를 successed 에 대입한다.
			successed = stat.executeUpdate(sql);
			
		} catch (SQLException e) {
			successed = 0;
		} finally {
			try {
				if (stat != null)
					stat.close();
			} catch (SQLException e) {
				// 무시한다?!
			}
		}
		
		return successed == 1;
	}

	/**
	 * blob 인 경우의 업데이트를 작성한다. byte[] 배열을 대입한다.
	 * 이 메소드는 구상 클래스에서 사용한다.
	 * @param tableName
	 * @param field
	 * @param valueBytes
	 * @return
	 */
	protected boolean updateBlobField(String tableName, String field, byte[] valueBytes, String key) {
		PreparedStatement pstat = null;
		boolean successed = false;
		
		try {
			pstat = connection.prepareStatement("update " + tableName
					+ "set value = ? "
					+ "where var_name = ?");

			// UPDATE {tableName}
			// SET {field} = {valueBytes}
			// WHERE var_name = {field}
			//
			// 결과를 successed 에 대입한다.
			pstat.setBytes(1, valueBytes);
			pstat.setString(2, field);
			successed = pstat.execute();
		} catch (SQLException e) {
			successed = false;
		} finally {
			try {
				if (pstat != null)
					pstat.close();
			} catch (SQLException e) {
				// 무시한다?!
			}
		}
		
		return successed;
	}
}
